# based on https://github.com/foolwood/SiamMask - MIT license
try:
    from fastai.basic_train import LearnerCallback
    from fastai.basic_train import Learner
    import torch
    import torch.nn as nn
    import traceback
    from torch.autograd import Variable
    import math
    import torch.utils.model_zoo as model_zoo
    from collections import namedtuple
    import torch.nn.functional as F
    import numpy as np
    import cv2
    import copy

    HAS_FASTAI = True
except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FASTAI = False


class SiamMask(nn.Module):
    def __init__(self, anchors=None, o_sz=127, g_sz=127):
        super().__init__()
        self.anchors = anchors  # anchor_cfg
        self.anchor_num = len(self.anchors["ratios"]) * len(self.anchors["scales"])
        self.anchor = Anchors(anchors)
        self.features = None
        self.rpn_model = None
        self.mask_model = None
        self.o_sz = o_sz
        self.g_sz = g_sz
        self.upSample = nn.UpsamplingBilinear2d(size=[g_sz, g_sz])

        self.all_anchors = None

    def set_all_anchors(self, image_center, size):
        # cx,cy,w,h
        if not self.anchor.generate_all_anchors(image_center, size):
            return
        all_anchors = self.anchor.all_anchors[1]  # cx, cy, w, h
        self.all_anchors = torch.from_numpy(all_anchors).float().cuda()
        self.all_anchors = [self.all_anchors[i] for i in range(4)]

    def feature_extractor(self, x):
        return self.features(x)

    def rpn(self, template, search):
        pred_cls, pred_loc = self.rpn_model(template, search)
        return pred_cls, pred_loc

    def mask(self, template, search):
        pred_mask = self.mask_model(template, search)
        return pred_mask

    def _add_rpn_loss(
        self,
        label_cls,
        label_loc,
        lable_loc_weight,
        label_mask,
        label_mask_weight,
        rpn_pred_cls,
        rpn_pred_loc,
        rpn_pred_mask,
    ):
        rpn_loss_cls = select_cross_entropy_loss(rpn_pred_cls, label_cls)

        rpn_loss_loc = weight_l1_loss(rpn_pred_loc, label_loc, lable_loc_weight)

        rpn_loss_mask, iou_m, iou_5, iou_7 = select_mask_logistic_loss(
            rpn_pred_mask, label_mask, label_mask_weight
        )

        return rpn_loss_cls, rpn_loss_loc, rpn_loss_mask, iou_m, iou_5, iou_7

    def run(self, template, search, softmax=False):
        """
        run network
        """
        template_feature = self.feature_extractor(template)
        # print(template_feature)
        feature, search_feature = self.features.forward_all(search)
        rpn_pred_cls, rpn_pred_loc = self.rpn(template_feature, search_feature)
        corr_feature = self.mask_model.mask.forward_corr(
            template_feature, search_feature
        )
        rpn_pred_mask = self.refine_model(
            feature[0],
            feature[1],
            feature[2],
            corr_feature,
            pos=None,
            test=torch.empty([1]),
        )

        if softmax:
            rpn_pred_cls = self.softmax(rpn_pred_cls)
        return (
            rpn_pred_cls,
            rpn_pred_loc,
            rpn_pred_mask,
            template_feature,
            search_feature,
        )

    def softmax(self, cls):
        b, a2, h, w = cls.size()
        cls = cls.view(b, 2, a2 // 2, h, w)
        cls = cls.permute(0, 2, 3, 4, 1).contiguous()
        cls = F.log_softmax(cls, dim=4)
        return cls

    def forward(self, input):
        """
        :param input: dict of input with keys of:
                'template': [b, 3, h1, w1], input template image.
                'search': [b, 3, h2, w2], input search image.
                'label_cls':[b, max_num_gts, 5] or None(self.training==False),
                                     each gt contains x1,y1,x2,y2,class.
        :return: dict of loss, predict, accuracy
        """
        template = input["template"]
        search = input["search"]
        if self.training:
            label_cls = input["label_cls"]
            label_loc = input["label_loc"]
            lable_loc_weight = input["label_loc_weight"]
            label_mask = input["label_mask"]
            label_mask_weight = input["label_mask_weight"]

        (
            rpn_pred_cls,
            rpn_pred_loc,
            rpn_pred_mask,
            template_feature,
            search_feature,
        ) = self.run(template, search, softmax=self.training)

        # return 0
        outputs = dict()

        outputs["predict"] = [
            rpn_pred_loc,
            rpn_pred_cls,
            rpn_pred_mask,
            template_feature,
            search_feature,
        ]
        if self.training:
            (
                rpn_loss_cls,
                rpn_loss_loc,
                rpn_loss_mask,
                iou_acc_mean,
                iou_acc_5,
                iou_acc_7,
            ) = self._add_rpn_loss(
                label_cls,
                label_loc,
                lable_loc_weight,
                label_mask,
                label_mask_weight,
                rpn_pred_cls,
                rpn_pred_loc,
                rpn_pred_mask,
            )
            outputs["losses"] = [rpn_loss_cls, rpn_loss_loc, rpn_loss_mask]
            outputs["accuracy"] = [iou_acc_mean, iou_acc_5, iou_acc_7]

        return outputs

    def template(self, z):
        self.zf = self.feature_extractor(z)
        cls_kernel, loc_kernel = self.rpn_model.template(self.zf)
        return cls_kernel, loc_kernel

    def track(self, x, cls_kernel=None, loc_kernel=None, softmax=False):
        xf = self.feature_extractor(x)
        rpn_pred_cls, rpn_pred_loc = self.rpn_model.track(xf, cls_kernel, loc_kernel)
        if softmax:
            rpn_pred_cls = self.softmax(rpn_pred_cls)
        return rpn_pred_cls, rpn_pred_loc


def get_cls_loss(pred, label, select):
    if select.nelement() == 0:
        return pred.sum() * 0.0
    pred = torch.index_select(pred, 0, select)
    label = torch.index_select(label, 0, select)

    return F.nll_loss(pred, label)


def select_cross_entropy_loss(pred, label):
    pred = pred.view(-1, 2)
    label = label.view(-1)
    pos = Variable(label.data.eq(1).nonzero().squeeze()).cuda()
    neg = Variable(label.data.eq(0).nonzero().squeeze()).cuda()

    loss_pos = get_cls_loss(pred, label, pos)
    loss_neg = get_cls_loss(pred, label, neg)
    return loss_pos * 0.5 + loss_neg * 0.5


def weight_l1_loss(pred_loc, label_loc, loss_weight):
    """
    :param pred_loc: [b, 4k, h, w]
    :param label_loc: [b, 4k, h, w]
    :param loss_weight:  [b, k, h, w]
    :return: loc loss value
    """
    b, _, sh, sw = pred_loc.size()
    pred_loc = pred_loc.view(b, 4, -1, sh, sw)
    diff = (pred_loc - label_loc).abs()
    diff = diff.sum(dim=1).view(b, -1, sh, sw)
    loss = diff * loss_weight
    return loss.sum().div(b)


def select_mask_logistic_loss(p_m, mask, weight, o_sz=63, g_sz=127):
    weight = weight.view(-1)
    pos = Variable(weight.data.eq(1).nonzero().squeeze())
    if pos.nelement() == 0:
        return p_m.sum() * 0, p_m.sum() * 0, p_m.sum() * 0, p_m.sum() * 0

    if len(p_m.shape) == 4:
        p_m = p_m.permute(0, 2, 3, 1).contiguous().view(-1, 1, o_sz, o_sz)
        p_m = torch.index_select(p_m, 0, pos)
        p_m = nn.UpsamplingBilinear2d(size=[g_sz, g_sz])(p_m)
        p_m = p_m.view(-1, g_sz * g_sz)
    else:
        p_m = torch.index_select(p_m, 0, pos)

    mask_uf = F.unfold(mask, (g_sz, g_sz), padding=0, stride=8)
    mask_uf = torch.transpose(mask_uf, 1, 2).contiguous().view(-1, g_sz * g_sz)

    mask_uf = torch.index_select(mask_uf, 0, pos)
    loss = F.soft_margin_loss(p_m, mask_uf)
    iou_m, iou_5, iou_7 = iou_measure(p_m, mask_uf)
    return loss, iou_m, iou_5, iou_7


def iou_measure(pred, label):
    pred = pred.ge(0).int()
    mask_sum = (pred.eq(1).int()).add(label.eq(1).int())
    intxn = torch.sum(mask_sum == 2, dim=1).float()
    union = torch.sum(mask_sum > 0, dim=1).float()
    iou = intxn / union

    return (
        torch.mean(iou),
        (torch.sum(iou > 0.5).float() / iou.shape[0]),
        (torch.sum(iou > 0.7).float() / iou.shape[0]),
    )


__all__ = ["ResNet", "resnet50"]

model_urls = {
    "resnet50": "https://download.pytorch.org/models/resnet50-19c8e357.pth",
}


def conv3x3(in_planes, out_planes, stride=1):
    "3x3 convolution with padding"
    return nn.Conv2d(
        in_planes, out_planes, kernel_size=3, stride=stride, padding=1, bias=False
    )


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = nn.BatchNorm2d(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = nn.BatchNorm2d(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class Features(nn.Module):
    def __init__(self):
        super(Features, self).__init__()
        self.feature_size = -1

    def forward(self, x):
        raise NotImplementedError

    def param_groups(self, start_lr, feature_mult=1):
        params = filter(lambda x: x.requires_grad, self.parameters())
        params = [{"params": params, "lr": start_lr * feature_mult}]
        return params

    def load_model(self, f="pretrain.model"):
        with open(f) as f:
            pretrained_dict = torch.load(f)
            model_dict = self.state_dict()
            pretrained_dict = {
                k: v for k, v in pretrained_dict.items() if k in model_dict
            }
            model_dict.update(pretrained_dict)
            self.load_state_dict(model_dict)


class MultiStageFeature(Features):
    def __init__(self):
        super(MultiStageFeature, self).__init__()

        self.layers = []
        self.train_num = -1
        self.change_point = []
        self.train_nums = []

    def unfix(self, ratio=0.0):
        if self.train_num == -1:
            self.train_num = 0
            self.unlock()
            self.eval()
        for p, t in reversed(list(zip(self.change_point, self.train_nums))):
            if ratio >= p:
                if self.train_num != t:
                    self.train_num = t
                    self.unlock()
                    return True
                break
        return False

    def train_layers(self):
        return self.layers[: self.train_num]

    def unlock(self):
        for p in self.parameters():
            p.requires_grad = False

        for m in self.train_layers():
            for p in m.parameters():
                p.requires_grad = True

    def train(self, mode):
        self.training = mode
        if mode == False:
            super(MultiStageFeature, self).train(False)
        else:
            for m in self.train_layers():
                m.train(True)

        return self


class Bottleneck(Features):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None, dilation=1):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        # padding = (2 - stride) + (dilation // 2 - 1)
        padding = 2 - stride
        assert (
            stride == 1 or dilation == 1
        ), "stride and dilation must have one equals to zero at least"
        if dilation > 1:
            padding = dilation
        self.conv2 = nn.Conv2d(
            planes,
            planes,
            kernel_size=3,
            stride=stride,
            padding=padding,
            bias=False,
            dilation=dilation,
        )
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, planes * 4, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(planes * 4)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        if out.size() != residual.size():
            print(out.size(), residual.size())
        out += residual

        out = self.relu(out)

        return out


class Bottleneck_nop(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(Bottleneck_nop, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(
            planes, planes, kernel_size=3, stride=stride, padding=0, bias=False
        )
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, planes * 4, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(planes * 4)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        s = residual.size(3)
        residual = residual[:, :, 1 : s - 1, 1 : s - 1]

        out += residual
        out = self.relu(out)

        return out


class ResNet(nn.Module):
    def __init__(self, block, layers, layer4=False, layer3=False):
        self.inplanes = 64
        super(ResNet, self).__init__()
        self.conv1 = nn.Conv2d(
            3, 64, kernel_size=7, stride=2, padding=0, bias=False  # 3
        )
        self.bn1 = nn.BatchNorm2d(64)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(block, 128, layers[1], stride=2)  # 31x31, 15x15

        self.feature_size = 128 * block.expansion

        if layer3:
            self.layer3 = self._make_layer(
                block, 256, layers[2], stride=1, dilation=2
            )  # 15x15, 7x7
            self.feature_size = (256 + 128) * block.expansion
        else:
            self.layer3 = lambda x: x  # identity

        if layer4:
            self.layer4 = self._make_layer(
                block, 512, layers[3], stride=1, dilation=4
            )  # 7x7, 3x3
            self.feature_size = 512 * block.expansion
        else:
            self.layer4 = lambda x: x  # identity

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2.0 / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def _make_layer(self, block, planes, blocks, stride=1, dilation=1):
        downsample = None
        dd = dilation
        if stride != 1 or self.inplanes != planes * block.expansion:
            if stride == 1 and dilation == 1:
                downsample = nn.Sequential(
                    nn.Conv2d(
                        self.inplanes,
                        planes * block.expansion,
                        kernel_size=1,
                        stride=stride,
                        bias=False,
                    ),
                    nn.BatchNorm2d(planes * block.expansion),
                )
            else:
                if dilation > 1:
                    dd = dilation // 2
                    padding = dd
                else:
                    dd = 1
                    padding = 0
                downsample = nn.Sequential(
                    nn.Conv2d(
                        self.inplanes,
                        planes * block.expansion,
                        kernel_size=3,
                        stride=stride,
                        bias=False,
                        padding=padding,
                        dilation=dd,
                    ),
                    nn.BatchNorm2d(planes * block.expansion),
                )

        layers = []
        # layers.append(block(self.inplanes, planes, stride, downsample, dilation=dilation))
        layers.append(block(self.inplanes, planes, stride, downsample, dilation=dd))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes, dilation=dilation))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        p0 = self.relu(x)
        x = self.maxpool(p0)

        p1 = self.layer1(x)
        p2 = self.layer2(p1)
        p3 = self.layer3(p2)

        return p0, p1, p2, p3


class ResAdjust(nn.Module):
    def __init__(
        self, block=Bottleneck, out_channels=256, adjust_number=1, fuse_layers=[2, 3, 4]
    ):
        super(ResAdjust, self).__init__()
        self.fuse_layers = set(fuse_layers)

        if 2 in self.fuse_layers:
            self.layer2 = self._make_layer(block, 128, 1, out_channels, adjust_number)
        if 3 in self.fuse_layers:
            self.layer3 = self._make_layer(block, 256, 2, out_channels, adjust_number)
        if 4 in self.fuse_layers:
            self.layer4 = self._make_layer(block, 512, 4, out_channels, adjust_number)

        self.feature_size = out_channels * len(self.fuse_layers)

    def _make_layer(self, block, plances, dilation, out, number=1):
        layers = []

        for _ in range(number):
            layer = block(plances * block.expansion, plances, dilation=dilation)
            layers.append(layer)

        downsample = nn.Sequential(
            nn.Conv2d(
                plances * block.expansion, out, kernel_size=3, padding=1, bias=False
            ),
            nn.BatchNorm2d(out),
        )
        layers.append(downsample)

        return nn.Sequential(*layers)

    def forward(self, p2, p3, p4):
        outputs = []

        if 2 in self.fuse_layers:
            outputs.append(self.layer2(p2))
        if 3 in self.fuse_layers:
            outputs.append(self.layer3(p3))
        if 4 in self.fuse_layers:
            outputs.append(self.layer4(p4))
        # return torch.cat(outputs, 1)
        return outputs


def resnet50(pretrained=False, **kwargs):
    """Constructs a ResNet-50 model.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = ResNet(Bottleneck, [3, 4, 6, 3], **kwargs)
    if pretrained:
        try:
            model.load_state_dict(model_zoo.load_url(model_urls["resnet50"]))
        except:
            pass
    return model


Corner = namedtuple("Corner", "x1 y1 x2 y2")
BBox = Corner
Center = namedtuple("Center", "x y w h")


def corner2center(corner):
    """
    :param corner: Corner or np.array 4*N
    :return: Center or 4 np.array N
    """
    if isinstance(corner, Corner):
        x1, y1, x2, y2 = corner
        return Center((x1 + x2) * 0.5, (y1 + y2) * 0.5, (x2 - x1), (y2 - y1))
    else:
        x1, y1, x2, y2 = corner[0], corner[1], corner[2], corner[3]
        x = (x1 + x2) * 0.5
        y = (y1 + y2) * 0.5
        w = x2 - x1
        h = y2 - y1
        return x, y, w, h


def center2corner(center):
    """
    :param center: Center or np.array 4*N
    :return: Corner or np.array 4*N
    """
    if isinstance(center, Center):
        x, y, w, h = center
        return Corner(x - w * 0.5, y - h * 0.5, x + w * 0.5, y + h * 0.5)
    else:
        x, y, w, h = center[0], center[1], center[2], center[3]
        x1 = x - w * 0.5
        y1 = y - h * 0.5
        x2 = x + w * 0.5
        y2 = y + h * 0.5
        return x1, y1, x2, y2


def cxy_wh_2_rect(pos, sz):
    return np.array([pos[0] - sz[0] / 2, pos[1] - sz[1] / 2, sz[0], sz[1]])  # 0-index


def get_axis_aligned_bbox(region):
    nv = region.size
    if nv == 8:
        cx = np.mean(region[0::2])
        cy = np.mean(region[1::2])
        x1 = min(region[0::2])
        x2 = max(region[0::2])
        y1 = min(region[1::2])
        y2 = max(region[1::2])
        A1 = np.linalg.norm(region[0:2] - region[2:4]) * np.linalg.norm(
            region[2:4] - region[4:6]
        )
        A2 = (x2 - x1) * (y2 - y1)
        s = np.sqrt(A1 / A2)
        w = s * (x2 - x1) + 1
        h = s * (y2 - y1) + 1
    else:
        x = region[0]
        y = region[1]
        w = region[2]
        h = region[3]
        cx = x + w / 2
        cy = y + h / 2

    return cx, cy, w, h


def aug_apply(bbox, param, shape, inv=False, rd=False):
    """
    apply augmentation
    :param bbox: original bbox in image
    :param param: augmentation param, shift/scale
    :param shape: image shape, h, w, (c)
    :param inv: inverse
    :param rd: round bbox
    :return: bbox(, param)
        bbox: augmented bbox
        param: real augmentation param
    """
    if not inv:
        center = corner2center(bbox)
        original_center = center

        real_param = {}
        if "scale" in param:
            scale_x, scale_y = param["scale"]
            imh, imw = shape[:2]
            h, w = center.h, center.w

            scale_x = min(scale_x, float(imw) / w)
            scale_y = min(scale_y, float(imh) / h)

            # center.w *= scale_x
            # center.h *= scale_y
            center = Center(center.x, center.y, center.w * scale_x, center.h * scale_y)

        bbox = center2corner(center)

        if "shift" in param:
            tx, ty = param["shift"]
            x1, y1, x2, y2 = bbox
            imh, imw = shape[:2]

            tx = max(-x1, min(imw - 1 - x2, tx))
            ty = max(-y1, min(imh - 1 - y2, ty))

            bbox = Corner(x1 + tx, y1 + ty, x2 + tx, y2 + ty)

        if rd:
            bbox = Corner(*map(round, bbox))

        current_center = corner2center(bbox)

        real_param["scale"] = (
            current_center.w / original_center.w,
            current_center.h / original_center.h,
        )
        real_param["shift"] = (
            current_center.x - original_center.x,
            current_center.y - original_center.y,
        )

        return bbox, real_param
    else:
        if "scale" in param:
            scale_x, scale_y = param["scale"]
        else:
            scale_x, scale_y = 1.0, 1.0

        if "shift" in param:
            tx, ty = param["shift"]
        else:
            tx, ty = 0, 0

        center = corner2center(bbox)

        center = Center(
            center.x - tx, center.y - ty, center.w / scale_x, center.h / scale_y
        )

        return center2corner(center)


def IoU(rect1, rect2):
    # overlap
    x1, y1, x2, y2 = rect1[0], rect1[1], rect1[2], rect1[3]
    tx1, ty1, tx2, ty2 = rect2[0], rect2[1], rect2[2], rect2[3]

    xx1 = np.maximum(tx1, x1)
    yy1 = np.maximum(ty1, y1)
    xx2 = np.minimum(tx2, x2)
    yy2 = np.minimum(ty2, y2)

    ww = np.maximum(0, xx2 - xx1)
    hh = np.maximum(0, yy2 - yy1)

    area = (x2 - x1) * (y2 - y1)

    target_a = (tx2 - tx1) * (ty2 - ty1)

    inter = ww * hh
    overlap = inter / (area + target_a - inter)

    return overlap


class Anchors:
    def __init__(self, cfg):
        self.stride = 8
        self.ratios = [0.33, 0.5, 1, 2, 3]
        self.scales = [8]
        self.round_dight = 0
        self.image_center = 0
        self.size = 0
        self.anchor_density = 1

        self.__dict__.update(cfg)

        self.anchor_num = (
            len(self.scales) * len(self.ratios) * (self.anchor_density**2)
        )
        self.anchors = None  # in single position (anchor_num*4)
        self.all_anchors = None  # in all position 2*(4*anchor_num*h*w)
        self.generate_anchors()

    def generate_anchors(self):
        self.anchors = np.zeros((self.anchor_num, 4), dtype=np.float32)

        size = self.stride * self.stride
        count = 0
        anchors_offset = self.stride / self.anchor_density
        anchors_offset = np.arange(self.anchor_density) * anchors_offset
        anchors_offset = anchors_offset - np.mean(anchors_offset)
        x_offsets, y_offsets = np.meshgrid(anchors_offset, anchors_offset)

        for x_offset, y_offset in zip(x_offsets.flatten(), y_offsets.flatten()):
            for r in self.ratios:
                if self.round_dight > 0:
                    ws = round(math.sqrt(size * 1.0 / r), self.round_dight)
                    hs = round(ws * r, self.round_dight)
                else:
                    ws = int(math.sqrt(size * 1.0 / r))
                    hs = int(ws * r)

                for s in self.scales:
                    w = ws * s
                    h = hs * s
                    self.anchors[count][:] = [
                        -w * 0.5 + x_offset,
                        -h * 0.5 + y_offset,
                        w * 0.5 + x_offset,
                        h * 0.5 + y_offset,
                    ][:]
                    count += 1

    def generate_all_anchors(self, im_c, size):
        if self.image_center == im_c and self.size == size:
            return False
        self.image_center = im_c
        self.size = size

        a0x = im_c - size // 2 * self.stride
        ori = np.array([a0x] * 4, dtype=np.float32)
        zero_anchors = self.anchors + ori

        x1 = zero_anchors[:, 0]
        y1 = zero_anchors[:, 1]
        x2 = zero_anchors[:, 2]
        y2 = zero_anchors[:, 3]

        x1, y1, x2, y2 = map(
            lambda x: x.reshape(self.anchor_num, 1, 1), [x1, y1, x2, y2]
        )
        cx, cy, w, h = corner2center([x1, y1, x2, y2])

        disp_x = np.arange(0, size).reshape(1, 1, -1) * self.stride
        disp_y = np.arange(0, size).reshape(1, -1, 1) * self.stride

        cx = cx + disp_x
        cy = cy + disp_y

        # broadcast
        zero = np.zeros((self.anchor_num, size, size), dtype=np.float32)
        cx, cy, w, h = map(lambda x: x + zero, [cx, cy, w, h])
        x1, y1, x2, y2 = center2corner([cx, cy, w, h])

        self.all_anchors = np.stack([x1, y1, x2, y2]), np.stack([cx, cy, w, h])
        return True


class Mask(nn.Module):
    def __init__(self):
        super(Mask, self).__init__()

    def forward(self, z_f, x_f):
        raise NotImplementedError

    def template(self, template):
        raise NotImplementedError

    def track(self, search):
        raise NotImplementedError

    def param_groups(self, start_lr, feature_mult=1):
        params = filter(lambda x: x.requires_grad, self.parameters())
        params = [{"params": params, "lr": start_lr * feature_mult}]
        return params


class RPN(nn.Module):
    def __init__(self):
        super(RPN, self).__init__()

    def forward(self, z_f, x_f):
        raise NotImplementedError

    def template(self, template):
        raise NotImplementedError

    def track(self, search):
        raise NotImplementedError

    def param_groups(self, start_lr, feature_mult=1, key=None):
        if key is None:
            params = filter(lambda x: x.requires_grad, self.parameters())
        else:
            params = [
                v for k, v in self.named_parameters() if (key in k) and v.requires_grad
            ]
        params = [{"params": params, "lr": start_lr * feature_mult}]
        return params


class DepthwiseConv2Group(nn.Module):
    def __init__(self):
        super(DepthwiseConv2Group, self).__init__()

    def forward(self, input, kernel):
        batch, channel = kernel.shape[:2]
        input = input.view(
            1, batch * channel, input.size(2), input.size(3)
        )  # 1 * (b*c) * k * k
        kernel = kernel.view(
            batch * channel, 1, kernel.size(2), kernel.size(3)
        )  # (b*c) * 1 * H * W
        feature = F.conv2d(input, kernel, groups=batch * channel)
        feature = feature.view(batch, channel, feature.size(2), feature.size(3))
        return feature


class DepthCorr(nn.Module):
    def __init__(self, in_channels, hidden, out_channels, kernel_size=3):
        super(DepthCorr, self).__init__()
        self.conv_kernel = nn.Sequential(
            nn.Conv2d(in_channels, hidden, kernel_size=kernel_size, bias=False),
            nn.BatchNorm2d(hidden),
            nn.ReLU(inplace=True),
        )
        self.conv_search = nn.Sequential(
            nn.Conv2d(in_channels, hidden, kernel_size=kernel_size, bias=False),
            nn.BatchNorm2d(hidden),
            nn.ReLU(inplace=True),
        )

        self.head = nn.Sequential(
            nn.Conv2d(hidden, hidden, kernel_size=1, bias=False),
            nn.BatchNorm2d(hidden),
            nn.ReLU(inplace=True),
            nn.Conv2d(hidden, out_channels, kernel_size=1),
        )

        self.dw_conv2d_group = DepthwiseConv2Group()

    def forward_corr(self, kernel, input):
        kernel = self.conv_kernel(kernel)
        input = self.conv_search(input)
        return self.dw_conv2d_group(input, kernel)

    def forward(self, kernel, search):
        feature = self.forward_corr(kernel, search)
        out = self.head(feature)
        return out

    def conv2d_dw_group(x, kernel):
        batch, channel = kernel.shape[:2]
        x = x.view(1, batch * channel, x.size(2), x.size(3))
        kernel = kernel.view(batch * channel, 1, kernel.size(2), kernel.size(3))
        out = F.conv2d(x, kernel, groups=batch * channel)
        out = out.view(batch, channel, out.size(2), out.size(3))
        return out


class ResDownS(nn.Module):
    def __init__(self, inplane, outplane):
        super(ResDownS, self).__init__()
        self.downsample = nn.Sequential(
            nn.Conv2d(inplane, outplane, kernel_size=1, bias=False),
            nn.BatchNorm2d(outplane),
        )

    def forward(self, x):
        x = self.downsample(x)
        if x.size(3) < 20:
            l = 4
            r = -4
            x = x[:, :, l:r, l:r]
        return x


class ResDown(MultiStageFeature):
    def __init__(self, pretrain=False, backbone="resnet50", pretrained_path=""):
        super(ResDown, self).__init__()
        if backbone == "resnet50":
            self.features = resnet50(layer3=True, layer4=False)

        if pretrain:
            load_pretrain(self.features, pretrained_path)

        self.downsample = ResDownS(1024, 256)

        self.layers = [self.downsample, self.features.layer2, self.features.layer3]
        self.train_nums = [1, 3]
        self.change_point = [0, 0.5]

        self.unfix(0.0)

    def param_groups(self, start_lr, feature_mult=1):
        lr = start_lr * feature_mult

        def _params(module, mult=1):
            params = list(filter(lambda x: x.requires_grad, module.parameters()))
            if len(params):
                return [{"params": params, "lr": lr * mult}]
            else:
                return []

        groups = []
        groups += _params(self.downsample)
        groups += _params(self.features, 0.1)
        return groups

    def forward(self, x):
        output = self.features(x)
        p3 = self.downsample(output[-1])
        return p3

    def forward_all(self, x):
        output = self.features(x)
        p3 = self.downsample(output[-1])
        return output, p3


class UP(RPN):
    def __init__(self, anchor_num=5, feature_in=256, feature_out=256):
        super(UP, self).__init__()

        self.anchor_num = anchor_num
        self.feature_in = feature_in
        self.feature_out = feature_out

        self.cls_output = 2 * self.anchor_num
        self.loc_output = 4 * self.anchor_num

        self.cls = DepthCorr(feature_in, feature_out, self.cls_output)
        self.loc = DepthCorr(feature_in, feature_out, self.loc_output)

    def forward(self, z_f, x_f):
        cls = self.cls(z_f, x_f)
        loc = self.loc(z_f, x_f)
        return cls, loc


class MaskCorr(Mask):
    def __init__(self, oSz=63):
        super(MaskCorr, self).__init__()
        self.oSz = oSz
        self.mask = DepthCorr(256, 256, self.oSz**2)

    def forward(self, z, x):
        return self.mask(z, x)


class Refine(nn.Module):
    def __init__(self):
        super(Refine, self).__init__()
        self.v0 = nn.Sequential(
            nn.Conv2d(64, 16, 3, padding=1),
            nn.ReLU(),
            nn.Conv2d(16, 4, 3, padding=1),
            nn.ReLU(),
        )

        self.v1 = nn.Sequential(
            nn.Conv2d(256, 64, 3, padding=1),
            nn.ReLU(),
            nn.Conv2d(64, 16, 3, padding=1),
            nn.ReLU(),
        )

        self.v2 = nn.Sequential(
            nn.Conv2d(512, 128, 3, padding=1),
            nn.ReLU(),
            nn.Conv2d(128, 32, 3, padding=1),
            nn.ReLU(),
        )

        self.h2 = nn.Sequential(
            nn.Conv2d(32, 32, 3, padding=1),
            nn.ReLU(),
            nn.Conv2d(32, 32, 3, padding=1),
            nn.ReLU(),
        )

        self.h1 = nn.Sequential(
            nn.Conv2d(16, 16, 3, padding=1),
            nn.ReLU(),
            nn.Conv2d(16, 16, 3, padding=1),
            nn.ReLU(),
        )

        self.h0 = nn.Sequential(
            nn.Conv2d(4, 4, 3, padding=1),
            nn.ReLU(),
            nn.Conv2d(4, 4, 3, padding=1),
            nn.ReLU(),
        )

        self.deconv = nn.ConvTranspose2d(256, 32, 15, 15)

        self.post0 = nn.Conv2d(32, 16, 3, padding=1)
        self.post1 = nn.Conv2d(16, 4, 3, padding=1)
        self.post2 = nn.Conv2d(4, 1, 3, padding=1)

        for modules in [
            self.v0,
            self.v1,
            self.v2,
            self.h2,
            self.h1,
            self.h0,
            self.deconv,
            self.post0,
            self.post1,
            self.post2,
        ]:
            for l in modules.modules():
                if isinstance(l, nn.Conv2d):
                    nn.init.kaiming_uniform_(l.weight, a=1)

    def forward(
        self, f0, f1, f2, corr_feature, pos=torch.empty(0), test=torch.empty(0)
    ):
        if torch.numel(test) == 0:
            p0 = torch.nn.functional.pad(f0, [16, 16, 16, 16])[
                :, :, 4 * pos[0] : 4 * pos[0] + 61, 4 * pos[1] : 4 * pos[1] + 61
            ]
            p1 = torch.nn.functional.pad(f1, [8, 8, 8, 8])[
                :, :, 2 * pos[0] : 2 * pos[0] + 31, 2 * pos[1] : 2 * pos[1] + 31
            ]
            p2 = torch.nn.functional.pad(f2, [4, 4, 4, 4])[
                :, :, pos[0] : pos[0] + 15, pos[1] : pos[1] + 15
            ]
        else:
            p0 = (
                F.unfold(f0, (61, 61), padding=0, stride=4)
                .permute(0, 2, 1)
                .contiguous()
                .view(-1, 64, 61, 61)
            )
            if not (pos is None):
                p0 = torch.index_select(p0, 0, pos[0])
            p1 = (
                F.unfold(f1, (31, 31), padding=0, stride=2)
                .permute(0, 2, 1)
                .contiguous()
                .view(-1, 256, 31, 31)
            )
            if not (pos is None):
                p1 = torch.index_select(p1, 0, pos[0])
            p2 = (
                F.unfold(f2, (15, 15), padding=0, stride=1)
                .permute(0, 2, 1)
                .contiguous()
                .view(-1, 512, 15, 15)
            )
            if not (pos is None):
                p2 = torch.index_select(p2, 0, pos[0])

        if not (pos is None):
            p3 = corr_feature[:, :, pos[0], pos[1]].view(-1, 256, 1, 1)
        else:
            p3 = corr_feature.permute(0, 2, 3, 1).contiguous().view(-1, 256, 1, 1)

        out = self.deconv(p3)
        out = self.post0(F.interpolate(self.h2(out) + self.v2(p2), size=(31, 31)))
        out = self.post1(F.interpolate(self.h1(out) + self.v1(p1), size=(61, 61)))
        out = self.post2(F.interpolate(self.h0(out) + self.v0(p0), size=(127, 127)))
        out = out.view(-1, 127 * 127)
        return out

    def param_groups(self, start_lr, feature_mult=1):
        params = filter(lambda x: x.requires_grad, self.parameters())
        params = [{"params": params, "lr": start_lr * feature_mult}]
        return params


class Custom(SiamMask):
    def __init__(
        self, pretrain=False, backbone="resnet50", pretrained_path="", **kwargs
    ):
        super().__init__(**kwargs)
        self.features = ResDown(
            pretrain=pretrain, backbone=backbone, pretrained_path=pretrained_path
        )
        self.rpn_model = UP(anchor_num=self.anchor_num, feature_in=256, feature_out=256)
        self.mask_model = MaskCorr()
        self.refine_model = Refine()

    def refine(self, f, pos=None):
        print("Refine function")
        return self.refine_model(f[0], f[1], f[2], pos)

    def template(self, template):
        self.zf = self.features(template)

    def track(self, search):
        search = self.features(search)
        rpn_pred_cls, rpn_pred_loc = self.rpn(self.zf, search)
        return rpn_pred_cls, rpn_pred_loc

    def track_mask(self, search):
        self.feature, self.search = self.features.forward_all(search)
        rpn_pred_cls, rpn_pred_loc = self.rpn(self.zf, self.search)
        self.corr_feature = self.mask_model.mask.forward_corr(self.zf, self.search)
        pred_mask = self.mask_model.mask.head(self.corr_feature)
        return rpn_pred_cls, rpn_pred_loc, pred_mask

    def track_refine(self, pos):
        pred_mask = self.refine_model(
            self.feature[0],
            self.feature[1],
            self.feature[2],
            self.corr_feature,
            pos=pos,
            test=torch.empty(0),
        )
        return pred_mask


def build_opt_lr(trainable_params):
    optimizer = torch.optim.SGD(
        trainable_params, 0.001, momentum=0.9, weight_decay=0.0001
    )

    return optimizer


def BNtoFixed(m):
    class_name = m.__class__.__name__
    if class_name.find("BatchNorm") != -1:
        m.eval()


def download_backbone(url, file_name):
    import os
    from pathlib import Path

    weights_path = os.path.join(Path.home(), ".cache", "siammask_checkpoint")
    if not os.path.exists(weights_path):
        os.makedirs(weights_path)
    weights_file = os.path.join(weights_path, file_name)
    if not os.path.exists(weights_file):
        try:
            import urllib.request

            print(
                f"[INFO] Downloading pretrained weights for SiamMask in {weights_file}"
            )
            urllib.request.urlretrieve(url, weights_file)
        except Exception as e:
            print(e)
            print("[INFO] Can't download pretrained weights for SiamMask.")
    return weights_file


def get_learner(data=None, anchors=None):
    if data is not None:
        file_path = download_backbone(
            url="http://www.robots.ox.ac.uk/~qwang/resnet.model",
            file_name="resnet.model",
        )
        model = Custom(
            anchors=anchors,
            pretrain=True,
            backbone="resnet50",
            pretrained_path=file_path,
        )
        model = model.cuda()

        learn = Learner(
            data=data,
            model=model,
            loss_func=siammask_loss,
            opt_func=build_opt_lr,
            metrics=mIOU,
        )

    return learn


def mIOU(outputs, *args):
    return outputs["accuracy"][0]


def siammask_loss(outputs, *args):
    rpn_cls_loss, rpn_loc_loss, rpn_mask_loss = (
        torch.mean(outputs["losses"][0]),
        torch.mean(outputs["losses"][1]),
        torch.mean(outputs["losses"][2]),
    )

    cls_weight, reg_weight, mask_weight = [1.0, 1.2, 36]

    loss = (
        rpn_cls_loss * cls_weight
        + rpn_loc_loss * reg_weight
        + rpn_mask_loss * mask_weight
    )

    return loss


class train_callback(LearnerCallback):
    def __init__(self, learn):
        super().__init__(learn)

    def on_batch_begin(self, last_input, last_target, **kwargs):
        "Handle new batch `xb`,`yb` in `train` or validation."
        x = {
            "template": torch.autograd.Variable(last_input[0]).cuda(),
            "search": torch.autograd.Variable(last_input[1]).cuda(),
            "label_cls": torch.autograd.Variable(last_input[2]).cuda(),
            "label_loc": torch.autograd.Variable(last_input[3]).cuda(),
            "label_loc_weight": torch.autograd.Variable(last_input[4]).cuda(),
            "label_mask": torch.autograd.Variable(last_input[6]).cuda(),
            "label_mask_weight": torch.autograd.Variable(last_input[7]).cuda(),
        }
        self.learn.model.train()
        last_target = torch.rand((4, 4, 4)).cuda()
        return {"last_input": x, "last_target": last_target}


def check_keys(model, pretrained_state_dict):
    ckpt_keys = set(pretrained_state_dict.keys())
    model_keys = set(model.state_dict().keys())
    used_pretrained_keys = model_keys & ckpt_keys
    assert len(used_pretrained_keys) > 0, "load NONE from pretrained checkpoint"
    return True


def remove_prefix(state_dict, prefix):
    """Old style model is stored with all names of parameters share common prefix 'module.'"""
    f = lambda x: x.split(prefix, 1)[-1] if x.startswith(prefix) else x
    return {f(key): value for key, value in state_dict.items()}


def load_pretrain(model, pretrained_path):
    if not torch.cuda.is_available():
        pretrained_dict = torch.load(
            pretrained_path, map_location=lambda storage, loc: storage
        )
    else:
        device = torch.cuda.current_device()
        pretrained_dict = torch.load(
            pretrained_path, map_location=lambda storage, loc: storage.cuda(device)
        )

    if "state_dict" in pretrained_dict.keys():
        pretrained_dict = remove_prefix(pretrained_dict["state_dict"], "module.")
    else:
        pretrained_dict = remove_prefix(pretrained_dict, "module.")

    try:
        check_keys(model, pretrained_dict)
    except:
        new_dict = {}
        for k, v in pretrained_dict.items():
            k = "features." + k
            new_dict[k] = v
        pretrained_dict = new_dict
        check_keys(model, pretrained_dict)
    model.load_state_dict(pretrained_dict, strict=False)
    return model


def to_torch(ndarray):
    if type(ndarray).__module__ == "numpy":
        return torch.from_numpy(ndarray)
    elif not torch.is_tensor(ndarray):
        raise ValueError("Cannot convert {} to torch tensor".format(type(ndarray)))
    return ndarray


def im_to_torch(img):
    img = np.transpose(img, (2, 0, 1))
    img = to_torch(img).float()
    return img


class TrackerConfig(object):
    penalty_k = 0.09
    window_influence = 0.39
    lr = 0.38
    seg_thr = 0.35
    windowing = "cosine"
    exemplar_size = 127
    instance_size = 255
    total_stride = 8
    out_size = 127
    base_size = 8
    score_size = (instance_size - exemplar_size) // total_stride + 1 + base_size
    context_amount = 0.5
    ratios = [0.33, 0.5, 1, 2, 3]
    scales = [
        8,
    ]
    anchor_num = len(ratios) * len(scales)
    round_dight = 0
    anchor = []

    def update(self, newparam=None, anchors=None):
        if newparam:
            for key, value in newparam.items():
                setattr(self, key, value)
        if anchors is not None:
            if isinstance(anchors, dict):
                anchors = Anchors(anchors)
            if isinstance(anchors, Anchors):
                self.total_stride = anchors.stride
                self.ratios = anchors.ratios
                self.scales = anchors.scales
                self.round_dight = anchors.round_dight
        self.renew()

    def renew(self):
        self.score_size = (
            (self.instance_size - self.exemplar_size) // self.total_stride
            + 1
            + self.base_size
        )
        self.anchor_num = len(self.ratios) * len(self.scales)


def get_subwindow_tracking(im, pos, model_sz, original_sz, avg_chans, out_mode="torch"):
    if isinstance(pos, float):
        pos = [pos, pos]
    sz = original_sz
    im_sz = im.shape
    c = (original_sz + 1) / 2
    context_xmin = round(pos[0] - c)
    context_xmax = context_xmin + sz - 1
    context_ymin = round(pos[1] - c)
    context_ymax = context_ymin + sz - 1
    left_pad = int(max(0.0, -context_xmin))
    top_pad = int(max(0.0, -context_ymin))
    right_pad = int(max(0.0, context_xmax - im_sz[1] + 1))
    bottom_pad = int(max(0.0, context_ymax - im_sz[0] + 1))

    context_xmin = context_xmin + left_pad
    context_xmax = context_xmax + left_pad
    context_ymin = context_ymin + top_pad
    context_ymax = context_ymax + top_pad

    r, c, k = im.shape
    if any([top_pad, bottom_pad, left_pad, right_pad]):
        te_im = np.zeros(
            (r + top_pad + bottom_pad, c + left_pad + right_pad, k), np.uint8
        )
        te_im[top_pad : top_pad + r, left_pad : left_pad + c, :] = im
        if top_pad:
            te_im[0:top_pad, left_pad : left_pad + c, :] = avg_chans
        if bottom_pad:
            te_im[r + top_pad :, left_pad : left_pad + c, :] = avg_chans
        if left_pad:
            te_im[:, 0:left_pad, :] = avg_chans
        if right_pad:
            te_im[:, c + left_pad :, :] = avg_chans
        im_patch_original = te_im[
            int(context_ymin) : int(context_ymax + 1),
            int(context_xmin) : int(context_xmax + 1),
            :,
        ]
    else:
        im_patch_original = im[
            int(context_ymin) : int(context_ymax + 1),
            int(context_xmin) : int(context_xmax + 1),
            :,
        ]

    if not np.array_equal(model_sz, original_sz):
        im_patch = cv2.resize(im_patch_original, (model_sz, model_sz))
    else:
        im_patch = im_patch_original

    return im_to_torch(im_patch) if out_mode in "torch" else im_patch


def generate_anchor(cfg, score_size):
    anchors = Anchors(cfg)
    anchor = anchors.anchors
    x1, y1, x2, y2 = anchor[:, 0], anchor[:, 1], anchor[:, 2], anchor[:, 3]
    anchor = np.stack([(x1 + x2) * 0.5, (y1 + y2) * 0.5, x2 - x1, y2 - y1], 1)

    total_stride = anchors.stride
    anchor_num = anchor.shape[0]

    anchor = np.tile(anchor, score_size * score_size).reshape((-1, 4))
    ori = -(score_size // 2) * total_stride
    xx, yy = np.meshgrid(
        [ori + total_stride * dx for dx in range(score_size)],
        [ori + total_stride * dy for dy in range(score_size)],
    )
    xx, yy = (
        np.tile(xx.flatten(), (anchor_num, 1)).flatten(),
        np.tile(yy.flatten(), (anchor_num, 1)).flatten(),
    )
    anchor[:, 0], anchor[:, 1] = xx.astype(np.float32), yy.astype(np.float32)
    return anchor


def siamese_init(im, target_pos, target_sz, model, hp=None, device="cpu"):
    state = dict()
    state["im_h"] = im.shape[0]
    state["im_w"] = im.shape[1]
    p = TrackerConfig()
    p.update(hp, model.anchors)

    p.renew()

    net = copy.deepcopy(model)
    p.scales = model.anchors["scales"]
    p.ratios = model.anchors["ratios"]
    p.anchor_num = model.anchor_num
    p.anchor = generate_anchor(model.anchors, p.score_size)
    avg_chans = np.mean(im, axis=(0, 1))

    wc_z = target_sz[0] + p.context_amount * sum(target_sz)
    hc_z = target_sz[1] + p.context_amount * sum(target_sz)
    s_z = round(np.sqrt(wc_z * hc_z))
    z_crop = get_subwindow_tracking(im, target_pos, p.exemplar_size, s_z, avg_chans)

    z = Variable(z_crop.unsqueeze(0))
    net.template(z.to(device))

    if p.windowing == "cosine":
        window = np.outer(np.hanning(p.score_size), np.hanning(p.score_size))
    elif p.windowing == "uniform":
        window = np.ones((p.score_size, p.score_size))
    window = np.tile(window.flatten(), p.anchor_num)

    state["p"] = p
    state["net"] = net
    state["avg_chans"] = avg_chans
    state["window"] = window
    state["target_pos"] = target_pos
    state["target_sz"] = target_sz
    return state


def siamese_track(
    state, im, mask_enable=False, refine_enable=False, device="cpu", debug=False
):
    p = state["p"]
    net = state["net"]
    avg_chans = state["avg_chans"]
    window = state["window"]
    target_pos = state["target_pos"]
    target_sz = state["target_sz"]

    wc_x = target_sz[1] + p.context_amount * sum(target_sz)
    hc_x = target_sz[0] + p.context_amount * sum(target_sz)
    s_x = np.sqrt(wc_x * hc_x)
    scale_x = p.exemplar_size / s_x
    d_search = (p.instance_size - p.exemplar_size) / 2
    pad = d_search / scale_x
    s_x = s_x + 2 * pad
    crop_box = [
        target_pos[0] - round(s_x) / 2,
        target_pos[1] - round(s_x) / 2,
        round(s_x),
        round(s_x),
    ]

    x_crop = Variable(
        get_subwindow_tracking(
            im, target_pos, p.instance_size, round(s_x), avg_chans
        ).unsqueeze(0)
    )

    if mask_enable:
        score, delta, mask = net.track_mask(x_crop.to(device))
    else:
        score, delta = net.track(x_crop.to(device))

    delta = delta.permute(1, 2, 3, 0).contiguous().view(4, -1).data.cpu().numpy()
    score = (
        F.softmax(
            score.permute(1, 2, 3, 0).contiguous().view(2, -1).permute(1, 0), dim=1
        )
        .data[:, 1]
        .cpu()
        .numpy()
    )

    delta[0, :] = delta[0, :] * p.anchor[:, 2] + p.anchor[:, 0]
    delta[1, :] = delta[1, :] * p.anchor[:, 3] + p.anchor[:, 1]
    delta[2, :] = np.exp(delta[2, :]) * p.anchor[:, 2]
    delta[3, :] = np.exp(delta[3, :]) * p.anchor[:, 3]

    def change(r):
        return np.maximum(r, 1.0 / r)

    def sz(w, h):
        pad = (w + h) * 0.5
        sz2 = (w + pad) * (h + pad)
        return np.sqrt(sz2)

    def sz_wh(wh):
        pad = (wh[0] + wh[1]) * 0.5
        sz2 = (wh[0] + pad) * (wh[1] + pad)
        return np.sqrt(sz2)

    target_sz_in_crop = target_sz * scale_x
    s_c = change(sz(delta[2, :], delta[3, :]) / (sz_wh(target_sz_in_crop)))
    r_c = change(
        (target_sz_in_crop[0] / target_sz_in_crop[1]) / (delta[2, :] / delta[3, :])
    )

    penalty = np.exp(-(r_c * s_c - 1) * p.penalty_k)
    pscore = penalty * score

    pscore = pscore * (1 - p.window_influence) + window * p.window_influence
    best_pscore_id = np.argmax(pscore)

    pred_in_crop = delta[:, best_pscore_id] / scale_x
    lr = penalty[best_pscore_id] * score[best_pscore_id] * p.lr

    res_x = pred_in_crop[0] + target_pos[0]
    res_y = pred_in_crop[1] + target_pos[1]

    res_w = target_sz[0] * (1 - lr) + pred_in_crop[2] * lr
    res_h = target_sz[1] * (1 - lr) + pred_in_crop[3] * lr

    target_pos = np.array([res_x, res_y])
    target_sz = np.array([res_w, res_h])

    if mask_enable:
        best_pscore_id_mask = np.unravel_index(
            best_pscore_id, (5, p.score_size, p.score_size)
        )
        delta_x, delta_y = best_pscore_id_mask[2], best_pscore_id_mask[1]

        if refine_enable:
            mask = (
                net.track_refine((delta_y, delta_x))
                .to(device)
                .sigmoid()
                .squeeze()
                .view(p.out_size, p.out_size)
                .cpu()
                .data.numpy()
            )
        else:
            mask = (
                mask[0, :, delta_y, delta_x]
                .sigmoid()
                .squeeze()
                .view(p.out_size, p.out_size)
                .cpu()
                .data.numpy()
            )

        def crop_back(image, bbox, out_sz, padding=-1):
            a = (out_sz[0] - 1) / bbox[2]
            b = (out_sz[1] - 1) / bbox[3]
            c = -a * bbox[0]
            d = -b * bbox[1]
            mapping = np.array([[a, 0, c], [0, b, d]]).astype(float)
            crop = cv2.warpAffine(
                image,
                mapping,
                (out_sz[0], out_sz[1]),
                flags=cv2.INTER_LINEAR,
                borderMode=cv2.BORDER_CONSTANT,
                borderValue=padding,
            )
            return crop

        s = crop_box[2] / p.instance_size
        sub_box = [
            crop_box[0] + (delta_x - p.base_size / 2) * p.total_stride * s,
            crop_box[1] + (delta_y - p.base_size / 2) * p.total_stride * s,
            s * p.exemplar_size,
            s * p.exemplar_size,
        ]
        s = p.out_size / sub_box[2]
        back_box = [
            -sub_box[0] * s,
            -sub_box[1] * s,
            state["im_w"] * s,
            state["im_h"] * s,
        ]
        mask_in_img = crop_back(mask, back_box, (state["im_w"], state["im_h"]))

        target_mask = (mask_in_img > p.seg_thr).astype(np.uint8)
        if cv2.__version__[-5] == "4":
            contours, _ = cv2.findContours(
                target_mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE
            )
        else:
            _, contours, _ = cv2.findContours(
                target_mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE
            )
        cnt_area = [cv2.contourArea(cnt) for cnt in contours]
        if len(contours) != 0 and np.max(cnt_area) > 100:
            contour = contours[np.argmax(cnt_area)]
            polygon = contour.reshape(-1, 2)
            prbox = cv2.boxPoints(cv2.minAreaRect(polygon))
            rbox_in_img = prbox
        else:
            location = cxy_wh_2_rect(target_pos, target_sz)
            rbox_in_img = np.array(
                [
                    [location[0], location[1]],
                    [location[0] + location[2], location[1]],
                    [location[0] + location[2], location[1] + location[3]],
                    [location[0], location[1] + location[3]],
                ]
            )

    target_pos[0] = max(0, min(state["im_w"], target_pos[0]))
    target_pos[1] = max(0, min(state["im_h"], target_pos[1]))
    target_sz[0] = max(10, min(state["im_w"], target_sz[0]))
    target_sz[1] = max(10, min(state["im_h"], target_sz[1]))

    state["target_pos"] = target_pos
    state["target_sz"] = target_sz
    state["score"] = score[best_pscore_id]
    state["mask"] = mask_in_img if mask_enable else []
    state["ploygon"] = rbox_in_img if mask_enable else []
    return state


def calculate_iou(rect1, rect2):
    rect1_min_x = rect1[0]
    rect1_min_y = rect1[1]
    rect1_max_x = rect1[0] + rect1[2]
    rect1_max_y = rect1[1] + rect1[3]

    rect2_min_x = rect2[0]
    rect2_min_y = rect2[1]
    rect2_max_x = rect2[0] + rect2[2]
    rect2_max_y = rect2[1] + rect2[3]

    inter_width = min(rect1_max_x, rect2_max_x) - max(rect1_min_x, rect2_min_x)
    inter_height = min(rect1_max_y, rect2_max_y) - max(rect1_min_y, rect2_min_y)

    inter_area = 0

    if inter_width > 0 and inter_height > 0:
        inter_area = inter_width * inter_height

    rect1_area = (rect1_max_x - rect1_min_x) * (rect1_max_y - rect1_min_y)
    rect2_area = (rect2_max_x - rect2_min_x) * (rect2_max_y - rect2_min_y)

    union_area = rect1_area + rect2_area - inter_area

    epsilon = 1e-6
    overlap_ratio = 0

    if inter_area > epsilon and union_area > epsilon:
        overlap_ratio = inter_area / union_area

    return overlap_ratio



# Dask Mapping
"""
Mapping Holds the Plot function for creating a FeatureCollection JSON plus the render options
"""
import uuid
import json
import dask.dataframe as dd
from arcgis.widgets import MapView

###########################################################################
##  Helper Lambda
###########################################################################
_fn_method = lambda a, op, **kwargs: getattr(a, op)(**kwargs)


###########################################################################
def dask_plot(df, map_widget=None, renderer=None):
    """

    Plot draws the data on a web map. The user can describe in simple terms how to
    renderer spatial data using symbol.  To make the process simplier a pallette
    for which colors are drawn from can be used instead of explicit colors.


    ======================  =========================================================
    **Explicit Argument**   **Description**
    ----------------------  ---------------------------------------------------------
    df                      required Dask DataFrame. This is the data to map.
    ----------------------  ---------------------------------------------------------
    map_widget              optional WebMap object. This is the map to display the
                            data on.
    ----------------------  ---------------------------------------------------------
    renderer                Optional dict-like.  The renderer definition for the dataset.
    ======================  =========================================================



    """
    renderer = None
    name = None
    map_exists = True
    if not hasattr(df, "spatial"):
        raise ValueError("DataFrame must be spatially enabled.")

    if df.spatial.renderer:
        renderer = json.loads(df.spatial.renderer.json)
    if name is None:
        name = uuid.uuid4().hex[:7]
    if map_widget is None:
        map_exists = False
        map_widget = MapView()
    assert isinstance(df, dd.DataFrame)

    feature_collections = df.map_partitions(
        lambda part: _fn_method(part.spatial, "to_feature_collection", **{"name": name})
    ).compute()
    if len(feature_collections) == 1:
        if map_exists:
            feature_collections[0].layer["layerDefinition"]["drawingInfo"][
                "renderer"
            ] = renderer
            map_widget.add_layer(feature_collections[0], options={"title": name})
        else:
            feature_collections[0].layer["layerDefinition"]["drawingInfo"][
                "renderer"
            ] = renderer
            map_widget.add_layer(feature_collections[0], options={"title": name})
    else:
        main_fc = feature_collections[0]
        main_fc.layer["layerDefinition"]["drawingInfo"]["renderer"] = renderer
        for fc in feature_collections[1:]:
            main_fc.properties["featureSet"]["features"].extend(
                fc.properties["featureSet"]["features"]
            )
            # fc.layer['layerDefinition']['drawingInfo']['renderer'] = renderer
        if map_exists:
            map_widget.add_layer(main_fc, options={"title": name})
        else:
            map_widget.add_layer(main_fc, options={"title": name})
    if map_exists == False:
        return map_widget
    return True



from urllib.parse import parse_qs
from getpass import getpass

from requests_oauthlib import OAuth1, OAuth2
from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib import OAuth2Session

from requests.auth import AuthBase
import lxml.html

from ._schain import SupportMultiAuth
from ..tools._lazy import LazyLoader
from ..tools import parse_url, assemble_url

warnings = LazyLoader("warnings")
re = LazyLoader("re")
json = LazyLoader("json")
webbrowser = LazyLoader("webbrowser")
getpass = LazyLoader("getpass")
_dt = LazyLoader("datetime")
requests = LazyLoader("requests")


###########################################################################
class EsriOAuth2Auth(AuthBase, SupportMultiAuth):
    """
    Performs the OAuth Workflow for logging in to Enterprise
    """

    _token = None
    _token_url = None
    _client_id = None
    _client_secret = None
    _username = None
    _password = None
    _token_url = None
    _referer = None
    _expiration = None
    _create_time = None
    _refresh_token = None
    _invalid_token_urls = None

    # ----------------------------------------------------------------------
    def __init__(
        self,
        base_url: str,
        client_id: str,
        client_secret: str = None,
        username: str = None,
        password: str = None,
        referer: str = "http",
        expiration: int = 1440,
        proxies: dict = None,
        session: "Session" = None,
        **kwargs,
    ) -> None:
        """
        Initializer for Oauth2 Workflows
        """
        self._invalid_token_urls = set()
        self.baseurl = base_url
        self._client_id = client_id
        self._client_secret = client_secret
        self.legacy = kwargs.pop("legacy", False)
        self._username = username
        if self._username and password is None:
            password = getpass.getpass(f"Enter user {username} password:")
        self._password = password
        if session is None:
            self._session = requests.Session()
            self._session.headers["referer"] = referer
            self._session.verify = kwargs.pop("veriy", True)
        else:
            self._session = session
        if proxies:
            self._proxies = proxies
        else:
            self._proxies = proxies

    # ----------------------------------------------------------------------
    def __str__(self):
        return f"<{self.__class__.__name__}, token=.....>"

    # ----------------------------------------------------------------------
    def __repr__(self):
        return f"<{self.__class__.__name__}, token=.....>"

    # ----------------------------------------------------------------------
    def _oauth_token(self):
        """performs the oauth2 when secret and client exist"""
        auth_url = "%s/oauth2/authorize" % self.baseurl
        tu = "%s/oauth2/token" % self.baseurl
        # handles the refreshing of the token
        if not (self._create_time is None) and (
            _dt.datetime.now()
            >= self._create_time + _dt.timedelta(minutes=self._expiration)
        ):
            self._token = None
        elif (
            not (self._create_time is None)
            and not (self._token is None)
            and (
                _dt.datetime.now()
                < self._create_time + _dt.timedelta(minutes=self._expiration)
            )
        ):
            return self._token
        # Handles token generation
        if (
            self._refresh_token is not None
            and self._client_id is not None
            and self._token is None
        ):  # Case 1: Refreshing a token
            parameters = {
                "client_id": self._client_id,
                "grant_type": "refresh_token",
                "refresh_token": self._refresh_token,
                "redirect_uri": "urn:ietf:wg:oauth:2.0:oob",
            }
            token_info = self._session.post(tu, data=parameters)
            self._token = token_info["access_token"]
            return self._token
        elif (
            self._client_id
            and self._client_secret
            and self._username
            and self._password
        ):
            oauth = OAuth2Session(
                client=BackendApplicationClient(client_id=self._client_id),
            )

            oauth.verify = False
            if self._proxies:
                oauth.proxies = self._proxies
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                res = oauth.fetch_token(
                    token_url=tu,
                    username=self._username,
                    password=self._password,
                    client_id=self._client_id,
                    client_secret=self._client_secret,
                    include_client_id=True,
                    verify=False,
                    proxies=self._proxies,
                    expiration=26000,
                )
            if "expires_in" in res:
                self._create_time = _dt.datetime.fromtimestamp(
                    res["expires_at"]
                ) - _dt.timedelta(seconds=7200)
                self._expiration = res["expires_in"] / 60
                if "token" in res:
                    return res["token"]
                if "access_token" in res:
                    return res["access_token"]
        elif (
            self._client_id and self._client_secret
        ):  # case 2: has both client and secret keys
            client = BackendApplicationClient(client_id=self._client_id)
            oauth = OAuth2Session(
                client=client, redirect_uri="urn:ietf:wg:oauth:2.0:oob"
            )
            if self._proxies:
                oauth.proxies = self._proxies
            oauth.verify = False
            res = oauth.fetch_token(
                token_url=tu,
                client_id=self._client_id,
                client_secret=self._client_secret,
                include_client_id=True,
                verify=False,
                proxies=self._proxies,
            )
            if "expires_in" in res:
                self._create_time = _dt.datetime.fromtimestamp(
                    res["expires_at"]
                ) - _dt.timedelta(seconds=7200)
                self._expiration = res["expires_in"] / 60
                if "token" in res:
                    return res["token"]
                if "access_token" in res:
                    return res["access_token"]
        elif (
            self._client_id and self._username is None and self._password is None
        ):  # case 3: client id only
            auth_url = "%s/oauth2/authorize" % self.baseurl
            tu = "%s/oauth2/token" % self.baseurl
            oauth = OAuth2Session(
                self._client_id, redirect_uri="urn:ietf:wg:oauth:2.0:oob"
            )
            if self._proxies:
                oauth.proxies = self._proxies
            oauth.verify = False
            authorization_url, state = oauth.authorization_url(
                auth_url, **{"allow_verification": "false"}
            )
            print(
                "Please sign in to your GIS and paste the code that is obtained below."
            )
            print(
                "If a web browser does not automatically open, please navigate to the URL below yourself instead."
            )
            print("Opening web browser to navigate to: " + authorization_url)

            webbrowser.open_new(authorization_url)
            authorization_response = getpass.getpass(
                "Enter code obtained on signing in using SAML: "
            )

            self._create_time = _dt.datetime.now()
            token_info = oauth.fetch_token(
                tu,
                code=authorization_response,
                verify=False,
                proxies=self._proxies,
                include_client_id=True,
                authorization_response="authorization_code",
            )
            self._expiration = token_info["expires_in"] / 60 - 2
            self._refresh_token = token_info["refresh_token"]
            self._token = token_info["access_token"]
            return self._token
        elif self._client_id and not (
            self._username is None and self._password is None
        ):  # case 4: client id and username/password (SAML workflow)
            parameters = {
                "client_id": self._client_id,
                "response_type": "code",
                "expiration": -1,  # we want refresh_token to work for the life of the script
                "redirect_uri": "urn:ietf:wg:oauth:2.0:oob",
                "allow_verification": "false",
            }
            content = str(self._session.get(auth_url, params=parameters).content)

            pattern = re.compile("var oAuthInfo = ({.*?});", re.DOTALL)
            if len(pattern.findall(content)) == 0:
                pattern = re.compile("var oAuthInfo = ({.*?})", re.DOTALL)

            soup = lxml.html.fromstring(content)

            def _load_oauth_info(js_object):
                """converts the js oauth to dict"""
                try:
                    oauth_info = json.loads(js_object)
                except:
                    oauth_info = json.loads(js_object + "}")
                return oauth_info

            for script in soup.xpath("//script/text()"):
                script_code = str(script).strip()
                matches = pattern.search(script_code)
                if not matches is None:
                    js_object = matches.groups()[0]
                    try:
                        oauth_info = _load_oauth_info(js_object)
                    except Exception:
                        raise Exception(
                            (
                                "Could not login. Please validate your creden"
                                "tials or make sure the security question is set on the user."
                            )
                        )
                    break

            parameters = {
                "user_orgkey": "",
                "username": self._username,
                "password": self._password,
                "oauth_state": oauth_info["oauth_state"],
            }
            resp = self._session.post(
                "%s/oauth2/signin" % self.baseurl,
                data=parameters,
                verify=False,
                proxies=self._proxies,
                allow_redirects=False,
            )
            if resp.status_code == 302:
                url = resp.headers["Location"]
                if url.find("acceptTermsAndConditions") > -1:
                    r2 = self._session.post(
                        url, data={"acceptTermsAndConditions": True}
                    )
                    content = r2.text
                elif url.find("oauth2/approval") > -1:
                    r2 = self._session.get(url)
                    content = r2.text

            soup = lxml.html.fromstring(content)
            codes = [
                t[len("SUCCESS code=") :]
                for t in soup.xpath("//title//text()")
                if t.find("SUCCESS") > -1
            ]
            if len(codes) > 0:
                code = codes[0]

            oauth = OAuth2Session(
                self._client_id, redirect_uri="urn:ietf:wg:oauth:2.0:oob"
            )
            if code is None:
                raise Exception("Could not generate a token.")
            self._create_time = _dt.datetime.now()
            token_info = oauth.fetch_token(
                tu,
                code=code,
                verify=False,
                include_client_id=True,
                authorization_response="authorization_code",
            )
            self._refresh_token = token_info["refresh_token"]
            self._token = token_info["access_token"]
            self._expiration = token_info["expires_in"] / 60 - 2

            return self._token
        return None

    # ----------------------------------------------------------------------
    @property
    def token(self) -> str:
        """
        Gets the Oauth token

        :returns: String
        """
        return self._oauth_token()

    # ----------------------------------------------------------------------
    def handle_40x(self, r, **kwargs):
        """Handles Case where token is invalid"""

        if (r.status_code < 500 and r.status_code > 399) or str(r.text).lower().find(
            "invalid token"
        ) > -1:
            # Recreate the request without the token
            #
            parsed = parse_url(r.url)
            server_url = assemble_url(parsed)
            self._invalid_token_urls.add(server_url)
            r.content
            r.raw.release_conn()
            r.request.headers.pop("X-Esri-Authorization", None)
            _r = r.connection.send(r.request, **kwargs)
            _r.headers["referer"] = self._referer or "http"
            _r.history.append(r)
            return _r
        return r

    # ----------------------------------------------------------------------
    def __call__(self, r):
        """Handles the Token Authorization Logic"""
        if self._invalid_token_urls is None:
            self._invalid_token_urls = set()
        parsed = parse_url(r.url)
        server_url = assemble_url(parsed)
        if not server_url in self._invalid_token_urls:
            r.register_hook("response", self.handle_40x)
            if self.legacy == False:
                r.headers["X-Esri-Authorization"] = f"Bearer {self._oauth_token()}"
                r.headers["referer"] = self._referer or ""
            elif self.legacy and r.method == "GET":
                r.prepare_url(url=r.url, params={"token": self._oauth_token()})
                r.headers["referer"] = self._referer or ""
            elif self.legacy and r.method == "POST":
                data = parse_qs(r.body)
                data["token"] = self._oauth_token()
                r.prepare_body(data, None, None)
                r.headers["referer"] = self._referer or ""
            else:
                raise Exception(
                    "Only POST and GET are supported with legacy methods of authentication."
                )
            try:
                self.pos = r.body.tell()
            except AttributeError:
                self.pos = None

            return r
        else:
            r.headers.pop("X-Esri-Authorization", None)
        return r



# Copyright 2018-2019 Open-MMLab.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# cascade_rcnn_r101_fpn_dconv_c3-c5_1x_coco.py, box AP=45.0

_base_ = "./_base_/models/cascade_rcnn_r50_fpn.py"

model = dict(
    pretrained="torchvision://resnet101",
    backbone=dict(
        dcn=dict(type="DCN", deform_groups=1, fallback_on_stride=False),
        stage_with_dcn=(False, True, True, True),
    ),
)

checkpoint = "http://download.openmmlab.com/mmdetection/v2.0/dcn/cascade_rcnn_r101_fpn_dconv_c3-c5_1x_coco/cascade_rcnn_r101_fpn_dconv_c3-c5_1x_coco_20200203-3b2f0594.pth"



import PIL

try:
    import os
    import torch
    import cv2
    from osgeo import gdal_array
    import numpy as np
    from torchvision.models import vgg16_bn
    import torch.nn.functional as F
    import torch.nn as nn
    from fastai.vision.learner import create_body
    from fastai.vision.models.unet import DynamicUnet
    from fastai.callbacks import hook_outputs
    from fastai.torch_core import requires_grad, children, apply_init
    from fastprogress.fastprogress import progress_bar
    from .._utils.superres import psnr, ssim
    from .._utils.common import ArcGISMSImage

    HAS_FASTAI = True
except Exception as e:
    # import_exception = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))

    HAS_FASTAI = False


def resize_to_new(img, targ_sz: int, use_min: bool = False):
    "Size to resize to, to hit `targ_sz` at same aspect ratio, in PIL coords (i.e w*h)"
    w, h, _ = img.shape
    min_sz = (min if use_min else max)(w, h)
    ratio = targ_sz / min_sz
    return int(w * ratio), int(h * ratio)


def resize_one(fn, i, path_lr, size, path_hr, img_size):
    dest = path_lr / fn.relative_to(path_hr)
    dest.parent.mkdir(parents=True, exist_ok=True)
    img = ArcGISMSImage.read_image(fn)
    targ_sz = resize_to_new(img, size, use_min=True)
    img = cv2.resize(img, targ_sz, interpolation=cv2.INTER_LINEAR)
    dest = str(dest.with_suffix(".tif"))
    if os.path.isfile(dest):
        os.remove(dest)
    ds = gdal_array.SaveArray(
        np.transpose(img, (2, 0, 1)), os.path.join(os.path.split(dest)[0], "test.tif")
    )
    ds = None
    os.rename(os.path.join(os.path.split(dest)[0], "test.tif"), dest)


def gram_matrix(x):
    n, c, h, w = x.size()
    x = x.view(n, c, -1)
    return (x @ x.transpose(1, 2)) / (c * h * w)


def UNetSR(data, arch, norm_type):
    from ._arcgis_model import _change_tail

    body = create_body(arch, pretrained=True)
    new_body = _change_tail(body, data)
    model = DynamicUnet(
        new_body,
        n_classes=data._n_channel,
        img_size=data.train_ds[0][0].size,
        blur=True,
        self_attention=True,
        norm_type=norm_type,
    )
    apply_init(model[2], nn.init.kaiming_normal_)
    return model


class FeatureLoss(nn.Module):
    def __init__(self, m_feat, layer_ids, layer_wgts, base_loss=F.l1_loss):
        super().__init__()
        self.m_feat = m_feat
        self.base_loss = base_loss
        self.loss_features = [self.m_feat[i] for i in layer_ids]
        self.hooks = hook_outputs(self.loss_features, detach=False)
        self.wgts = layer_wgts
        self.metric_names = [
            "pixel_loss",
        ]

    def make_features(self, x, clone=False):
        self.m_feat(x)
        return [(o.clone() if clone else o) for o in self.hooks.stored]

    def forward(self, input, target):
        out_feat = self.make_features(target, clone=True)
        in_feat = self.make_features(input)
        self.feat_losses = [self.base_loss(input, target)]
        self.feat_losses += [
            self.base_loss(f_in, f_out) * w
            for f_in, f_out, w in zip(in_feat, out_feat, self.wgts)
        ]
        self.feat_losses += [
            self.base_loss(gram_matrix(f_in), gram_matrix(f_out)) * w**2 * 5e3
            for f_in, f_out, w in zip(in_feat, out_feat, self.wgts)
        ]
        self.metrics = dict(zip(self.metric_names, self.feat_losses))
        return sum(self.feat_losses)

    def __del__(self):
        self.hooks.remove()


def create_loss(c, device_type="cuda"):
    if device_type == "cuda":
        device = torch.device("cuda")
    else:
        device = torch.device("cpu")
    vggmodel = vgg16_bn(True)
    vggmodel.features[0] = nn.Conv2d(c, 64, 3, 1, 1)
    vgg_m = vggmodel.features.to(device).eval()
    requires_grad(vgg_m, False)
    blocks = [
        i - 1 for i, o in enumerate(children(vgg_m)) if isinstance(o, nn.MaxPool2d)
    ]
    blocks, [vgg_m[i] for i in blocks]
    feat_loss = FeatureLoss(vgg_m, blocks[2:5], [5, 15, 2])
    return feat_loss


def compute_metrics(model, dl, show_progress, **kwargs):
    sampling = kwargs.get("sampling_type", "ddim")
    ntimestep = kwargs.get("n_timestep")

    avg_psnr = 0
    avg_ssim = 0
    model.learn.model.eval()
    with torch.no_grad():
        for input, target in progress_bar(dl, display=False):
            if model.model_type == "UNet":
                prediction = model.learn.model(input)
            else:
                device = next(model.learn.model.parameters()).device.type

                if sampling == "ddim":
                    n_timestep = ntimestep if ntimestep else 200
                    nstp = {"n_timestep": model.kwargs.get("n_timestep", 1000)}
                else:
                    n_timestep = (
                        ntimestep if ntimestep else model.kwargs.get("n_timestep", 1000)
                    )
                    nstp = {"n_timestep": n_timestep}
                combkwargs = {**kwargs, **model.kwargs, **nstp}
                model.learn.model.set_new_noise_schedule(device, **combkwargs)

                from tqdm import tqdm

                preds = []
                for k in tqdm(
                    range(input[0].shape[0]), desc="sampling loop time step per image"
                ):
                    if sampling == "ddim":
                        preds.append(
                            model.learn.model.super_resolution(
                                input[0][k, None],
                                continous=False,
                                sampling_timesteps=n_timestep,
                                ddim_sampling_eta=1,
                                sampling="ddim",
                            )
                        )
                    else:
                        preds.append(
                            model.learn.model.super_resolution(
                                input[0][k, None], continous=False
                            )
                        )
                prediction = torch.cat(preds)

                avg_psnr += psnr(prediction, target)
                avg_ssim += ssim(prediction, target)
                break
            avg_psnr += psnr(prediction, target)
            avg_ssim += ssim(prediction, target)
    if model.model_type == "UNet":
        return avg_psnr / len(dl), avg_ssim.item() / len(dl)
    else:
        return avg_psnr, avg_ssim


def get_resize(y, z, max_size, f):
    if y * f <= max_size and z * f <= max_size:
        y_new = y * f
        z_new = z * f
    else:
        if y > z:
            y_new = max_size
            z_new = int(round_up_to_even(z * max_size / y))
        else:
            z_new = max_size
            y_new = int(round_up_to_even(y * max_size / z))
    return (y_new, z_new)



try:
    from ._data import _raise_fastai_import_error
    from ._tracking.track_processor import TrackProcessor, get_default_tracker_options
    import traceback

    HAS_FASTAI = True
except ImportError:
    import_exception = traceback.format_exc()
    HAS_FASTAI = False


class ObjectTracker:
    """
    Creates :class:`~arcgis.learn.ObjectTracker` Object.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    tracker                 Required. Returned tracker object from
                            from_model API of object tracking models.
    ---------------------   -------------------------------------------
    detector                Optional. Returned detector object from
                            from_model API of object detection models.
    ---------------------   -------------------------------------------
    tracker_options         Optional dictionary. A dictionary with
                            keys as parameter names and values as
                            parameter values.

                            * "``enable_post_processing``" - refers to
                              the flag which enables/disables post_processing
                              of tracks internal to ObjectTracker module.
                              For DeepSort, it's recommended to keep this
                              flag as False. Default - True

                            * "``detection_interval``" - refers to
                              the interval in frames at which the detector
                              is invoked. It should be >= 1

                            * "``detection_threshold``" - refers to
                              the lower threshold for selecting the
                              detections.

                            * "``detect_track_failure``" - refers to
                              the flag which enables/disables the logic
                              to detect whether the object appearance
                              has changed detection.

                            * "``recover_track``" - refers to the flag which
                              enables/disables track recovery post failure.

                            * "``stab_period``" - refers to the number of frames
                              after which post processing starts.

                            * "``detect_fail_interval``" - refers to the number
                              of frames after which to detect track failure.

                            * "``min_obj_size``" - refers to the size in pixels
                              below which tracking is assumed to have
                              failed.

                            * "``template_history``" - refers to the number of
                              frames before the current frame at which
                              template image is fetched.

                            * "``status_history``" - refers to the
                              number of frames over which status of the
                              track is used to detect track failure.

                            * "``status_fail_threshold``" - refers to the
                              threshold for the ratio between number
                              of frames for which object is searched
                              for and the total number of frames which
                              needs to be crossed for track failure
                              detection.

                            * "``search_period``" - refers to the
                              number of frames for which object is
                              searched for before declaring object is
                              lost.

                            * "``knn_distance_ratio``" - refers to the
                              threshold for ratio of the distances between
                              template descriptor and the two best matched
                              detection descriptor, used for filtering
                              best matches.

                            * "``recover_conf_threshold``" -  refers
                              to the minimum confidence value over which
                              recovery logic is enabled.

                            * ``recover_iou_threshold`` - refers to the minimum
                              overlap between template and detection for
                              successful recovery.
    =====================   ===========================================

    :return: :class:`~arcgis.learn.ObjectTracker` Object
    """

    def __init__(
        self,
        tracker,
        detector=None,
        tracker_options={
            "enable_post_processing": True,
            "detection_interval": 5,
            "detection_threshold": 0.3,
            "detect_track_failure": True,
            "recover_track": True,
            "stab_period": 6,
            "detect_fail_interval": 5,
            "min_obj_size": 10,
            "template_history": 25,
            "status_history": 60,
            "status_fail_threshold": 0.6,
            "search_period": 60,
            "knn_distance_ratio": 0.75,
            "recover_conf_threshold": 0.1,
            "recover_iou_threshold": 0.1,
        },
    ):
        if not HAS_FASTAI:
            _raise_fastai_import_error(import_exception=import_exception)

        if tracker_options is None:
            tracker_options = get_default_tracker_options()

        self.tracker = tracker
        self.detector = detector
        self.tracks = []
        self.detect_interval = tracker_options.get("detection_interval", 10)
        self.detect_threshold = tracker_options.get("detection_threshold", 0.3)
        self.enable_post_processing = tracker_options.get(
            "enable_post_processing", True
        )
        processor_options = {}
        processor_options["detect_track_failure"] = tracker_options.get(
            "detect_track_failure", True
        )
        processor_options["recover_track"] = tracker_options.get("recover_track", True)
        processor_options["stab_period"] = tracker_options.get("stab_period", 6)
        processor_options["detect_fail_interval"] = tracker_options.get(
            "detect_fail_interval", 5
        )
        processor_options["min_obj_size"] = tracker_options.get("min_obj_size", 10)
        processor_options["template_history"] = tracker_options.get(
            "template_history", 25
        )
        processor_options["status_history"] = tracker_options.get("status_history", 60)
        processor_options["status_fail_threshold"] = tracker_options.get(
            "status_fail_threshold", 0.6
        )
        processor_options["search_period"] = tracker_options.get("search_period", 60)
        processor_options["knn_distance_ratio"] = tracker_options.get(
            "knn_distance_ratio", 0.75
        )
        processor_options["recover_conf_threshold"] = tracker_options.get(
            "recover_conf_threshold", 0.1
        )
        processor_options["recover_iou_threshold"] = tracker_options.get(
            "recover_iou_threshold", 0.1
        )

        self.processor = None
        if self.enable_post_processing:
            TrackProcessor(processor_options)
        self.frames_processed = 0

    def init(
        self, frame, detections=None, labels=None, reset=True
    ):  # TODO: pass scores
        """
        Initializes tracks based on the detections returned by detector/
        manually fed to the function.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        frame                   Required numpy array. frame is used to
                                initialize the objects to track.
        ---------------------   -------------------------------------------
        detections              Optional list. A list of bounding box to
                                intialize the tracks.
        ---------------------   -------------------------------------------
        labels                  Optional list. A list of labels corresponding
                                to the detections.
        ---------------------   -------------------------------------------
        reset                   Optional flag. Indicates whether to reset
                                the tracker and remove all existing tracks
                                before initialization.
        =====================   ===========================================

        :return: list of active track objects
        """
        if detections is None:
            if self.detector is not None:
                predictions, labels, scores = self.detector.predict(
                    frame, return_scores=True
                )
                detections, labels, scores = self._filter_dets(
                    predictions, labels, scores
                )
        else:
            # TODO: see deepsort get_corrected_labels_scores
            if labels is None or len(labels) != len(detections):
                labels = ["Object"] * len(detections)

        self.tracks = self.tracker.init(
            frame,
            detections=detections,
            labels=labels,
            reset=reset,
            scores=scores,
            update_interval=self.detect_interval,
        )

        if (
            self.enable_post_processing is True
            and self.processor is not None
            and not detections is None
        ):
            track_list = self._convert_tracks_to_list(self.tracks)
            self.processor.init(frame, track_list, reset)

        if reset:
            self.frames_processed = 0
        return self._get_active_tracks(self.tracks)

    def update(self, frame):
        """
        Tracks the position of the object in the frame/Image.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        frame                   Required numpy array. frame is the current
                                frame to be used to track the objects.
        =====================   ===========================================

        :return: list of active track objects
        """
        if (
            self.frames_processed % self.detect_interval == 0
            and self.detector is not None
        ):
            predictions, labels, scores = self.detector.predict(
                frame, return_scores=True
            )
            detections, labels, scores = self._filter_dets(predictions, labels, scores)
            self.tracks = self.tracker.update(
                frame, detections=detections, labels=labels, scores=scores
            )
            if self.enable_post_processing is True and self.processor is not None:
                tracks_list = self._convert_tracks_to_list(self.tracks)
                self.processor.init(frame, tracks_list, False)
        else:
            self.tracks = self.tracker.update(frame)
            if self.enable_post_processing is True and self.processor is not None:
                tracks_list = self._convert_tracks_to_list(self.tracks)
                tracks_list = self.processor.update(frame, tracks_list)
                self.tracks = self._update_processed_tracks(tracks_list)

        self.frames_processed = self.frames_processed + 1
        return self._get_active_tracks(self.tracks)

    def remove(self, tracks_ids):
        """
        Removes the tracks corresponding to track_ids parameter.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        tracks_ids              Required list. List of track ids to be
                                removed.
        =====================   ===========================================

        """
        self.tracks = self.tracker.remove(tracks_ids)
        if self.enable_post_processing is True and self.processor is not None:
            self.processor.remove(tracks_ids)

        return self.tracks

    def _filter_dets(self, predictions, labels, scores):
        """
        Filters the predictions and then converts it to 1D list which can be used by
        TrackProcessor.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        predictions             Required list. Predictions which satisfy
                                selection criteria are used.
        ---------------------   -------------------------------------------
        labels                  Optional list. A list containing labels for
                                the predictions.
        ---------------------   -------------------------------------------
        scores                  Required list. A list containing scores for
                                the predictions.
        =====================   ===========================================

        :return: 1D lists with predictions, labels, scores
        """
        if predictions is None or len(predictions) == 0:
            return [], [], []

        tdets = []
        tlabels = []
        tscores = []
        offset = 0
        max_objects = 1000

        for index in range(offset, min(max_objects, len(predictions) - offset)):
            prediction = predictions[index]
            label = "Object"
            score = 0.0
            if labels is not None and index < len(labels):
                label = labels[index]
            if scores is not None and index < len(scores):
                score = scores[index]

            if score > self.detect_threshold:
                tdets.append(prediction)
                tlabels.append(label)
                tscores.append(score)

        return tdets, tlabels, tscores

    def _convert_tracks_to_list(self, tracks):
        """
        Converts tracks list to 1D list which can be used by
        TrackProcessor.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        tracks                  Required list. List of tracks to be used by
                                TrackProcessor.
        =====================   ===========================================

        :return: 1D list with values of track members used by
                  TrackProcessor
        """
        tracks_list = []
        if tracks is None:
            return tracks_list

        tracks_list.append(len(tracks))
        tracks_list.append(10)

        for track in tracks:
            tracks_list.append(int(track.id))
            tracks_list.append(str(track.label))  # int(track.class_id)
            tracks_list.append(int(track.age))
            tracks_list.append(float(track.score))
            tracks_list.append(int(track.bbox[0]))
            tracks_list.append(int(track.bbox[1]))
            tracks_list.append(int(track.bbox[2]))
            tracks_list.append(int(track.bbox[3]))
            if track.mask is not None:
                tracks_list.append(255 * track.mask)
            else:
                tracks_list.append(track.mask)
            tracks_list.append(int(track.status))
        return tracks_list

    def _update_processed_tracks(self, tracks_list):
        """
        Uses 1D list returned by TrackProcessor to update list of tracks.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        tracks_list             Required list. 1D list with values of track
                                members.
        =====================   ===========================================

        :return: list of track objects
        """
        if tracks_list is None or len(tracks_list) == 0:
            return []

        num_dets = tracks_list[0]
        len_det = tracks_list[1]

        for i in range(0, num_dets):
            det_index = (i * len_det) + 2
            id = tracks_list[det_index]

            track_index = -1

            for j, track in enumerate(self.tracks):
                if id == track.id:
                    track_index = j
                    break

            if track_index >= 0:
                self.tracks[track_index].status = tracks_list[det_index + 9]

        return self.tracks

    def _get_active_tracks(self, tracks):
        """
        Filters the active tracks using the argument tracks.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        tracks                  Required list. List of tracks to be
                                filtered.
        =====================   ===========================================

        :return: list of active track objects
        """
        # TODO: 16
        active_tracks = list(filter(lambda track: track.status == 16, tracks))
        return active_tracks


# TODO: Implement predict_video



try:
    import arcgis as ag
    from fastai.basic_train import LearnerCallback
    import traceback
    import torch
    import matplotlib.pyplot as plt
    import pandas as pd
    import gc
    from IPython.display import clear_output
    from datetime import datetime as dt
    import io, os, base64
    import importlib
    import optuna
    import time
    import numpy as np

    from fastai.data_block import get_files
    from pathlib import Path
    from zipfile import ZipFile
    import tempfile

    HAS_FASTAI = True

    HAS_FASTAI = True
except Exception as e:
    print(e)
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FASTAI = False


class ToolIsCancelled(Exception):
    pass


self_obj = None
best_name_time = None
(
    all_val_losses,
    all_train_losses,
    dice,
    BestPerformingModel,
    best_backbone,
    best_model,
    timing,
) = ([], [], [], None, None, None, [])


class train_callback(LearnerCallback):
    def __init__(self, learn, stop_var):
        self.counter = 0
        self.stop_var = stop_var
        super().__init__(learn)

    def on_batch_end(self, **kwargs):
        # print(self.counter)
        self.counter += 1
        is_present = importlib.util.find_spec("arcpy")
        if is_present is not None:
            import arcpy

            if arcpy.env.isCancelled:
                raise ToolIsCancelled("Function aborted by User.")
        if self.counter > self.stop_var:
            self.counter = 0
            return {"stop_epoch": True}


def generate_output_report(
    df,
    output_folder,
    exhaustive_mode_studies,
    mode,
    save_to_folder,
    save_evaluated_models,
):
    content = ""
    header = """
            <!DOCTYPE html>
            <html>
            <head>
                <style>
                
            .styled-table {
                border-collapse: collapse;
                font-size: 0.9em;
                font-family:Courier New;
            }

            .styled-table td, .styled-table th {
                border: 1px solid #ddd;
                padding: 8px;
            }

            .styled-table tr:nth-child(even){background-color: #f2f2f2;}

            .styled-table tr:hover {background-color: #e0ecf5;}

            .styled-table thead {
                padding-top: 6px;
                padding-bottom: 6px;
                text-align: left;
                background-color: #0099cc;
                color: white;
            }

            body {
                font-family: Arial;
                font-size: 1.0em;
                background-color: rgba(236, 243, 249, 0.15);
            }

            h1 {
                color: #004666;
                border-bottom: 1px solid rgba(0,70,102,0.3)
            }
            h2 {
                color: #004666;
                padding-bottom: 5px;
                margin-bottom: 0px;
            }

            ul {
                margin-top: 0px;
            }

            p {
                margin-top: 5px;
            }

            h3 {
                color: #004666;
                padding-bottom: 5px;
                margin-bottom: 0px;
            }
            a {
                font-weight: bold;
                color: #004666;
            }

            a:hover {
                cursor: pointer;
                color: #0099CC;
            }



                </style>
            </head>
            <body>
                <div id="main">
                <h1>AutoDL Leaderboard</h1>

        """

    footer = r"""
            <script>
                function toggleShow(elementId) {
                    var x = document.getElementById(elementId);
                    if (x.style.display === "none") {
                        x.style.display = "block";
                    } else {
                        x.style.display = "none";
                    }
                }
            </script>
        </body>
        </html>
    """
    # print(df)
    data_x = list(
        pd.Series(df.index).astype(str) + "_" + df["Model"] + "_" + df["backbone"]
    )[::-1]
    if "accuracy" in df.keys():
        data_y = list(df["accuracy"])[::-1]
        accuracy_fn = "accuracy"
    else:
        data_y = list(df["average_precision_score"])[::-1]
        accuracy_fn = "average_precision_score"

    fig = plt.figure(figsize=(20, 10))

    # creating the bar plot
    bars = plt.barh(data_x, data_y, color="#5ca3d0")
    # addlabels(data_x, data_y)
    # print(bars)

    for bar in bars:
        plt.text(
            bar.get_x() + bar.get_width(),
            bar.get_y() + bar.get_height() / 2,
            round(bar.get_width(), 2),
            size=14,
        )

    plt.ylabel("Model Names")
    plt.xlabel("accuracy")

    plt.title("AutDL Performance")
    # plt.show()

    pic_IObytes = io.BytesIO()
    fig.savefig(pic_IObytes, format="png", dpi=400)
    pic_IObytes.seek(0)
    pic_hash = base64.b64encode(pic_IObytes.read())

    plt.close(fig)

    table = '<table class="styled-table">'

    table_head = f"<thead>\n<tr style='text-align: right;'>\n"
    if accuracy_fn == "average_precision_score":
        df_ = df[
            [
                "Model",
                "train_loss",
                "valid_loss",
                "average_precision_score",
                "lr",
                "training time",
                "optuna_study",
                "backbone",
            ]
        ]
    else:
        df_ = df[
            [
                "Model",
                "train_loss",
                "valid_loss",
                "accuracy",
                "dice",
                "lr",
                "training time",
                "optuna_study",
                "backbone",
            ]
        ]
    table_header_content = list(df_.keys())
    table_body_content = df_.values.tolist()
    # print(table_body_content)
    for thc in table_header_content:
        table_head += "<th style='text-align: left;'>" + thc + "</th>\n"

    table_head += "</tr>\n</thead>\n"

    table_content = "<tbody>"

    for i, tbc in enumerate(table_body_content, start=1):
        table_content += '<tr style="text-align: right;">\n'
        for j, data in enumerate(tbc, start=1):
            if j == 1:
                id_n = str(round(tbc[3], 6)).replace(".", "_")
                table_content += (
                    """<td style="text-align: left;"><u>
                            <a onclick="toggleShow(\'"""
                    + str(data)
                    + "_"
                    + tbc[-1]
                    + id_n
                    + """\');toggleShow('main')" >
                                """
                    + str(i)
                    + """_"""
                    + str(data)
                    + """
                            </u></a>
                        </td>\n"""
                )
            else:
                if j < len(tbc) - 2:
                    data = round(data, 6)
                table_content += (
                    """<td style="text-align: left;">
                                """
                    + str(data)
                    + """
                        </td>\n"""
                )

        table_content += "</tr>\n"
    table_content += "</tbody>\n"
    table = table + table_head + table_content + "</table> "

    content += table

    div_accuracy_chart = (
        """<h3>AutoDL model's accuracy graph</h3>
    <p><img style="width:1100px" alt="AutoDL Performance" src="data:image/png;base64,"""
        + pic_hash.decode("utf-8")
        + """" /></p>

    """
    )

    content += div_accuracy_chart

    # add model report

    if mode == "advanced":
        model_report = ""
        model_report += "<h2>Network wise study details</h2>"

        for i, ex in enumerate(exhaustive_mode_studies):
            model_report += "<hr/><h3>Network Name: " + ex.study_name + "</h3>"

            model_report += "<h4>Best parameter combination</h4>"
            best_params = ex.best_params
            for key, val in best_params.items():
                model_report += (
                    "<li><strong>" + str(key) + "</strong>: " + str(val) + "</li>"
                )

            try:
                img_details = optuna.importance.get_param_importances(ex)
                data_x = list(img_details.keys())
                data_y = list(img_details.values())

                fig = plt.figure(figsize=(12, 7))

                # creating the bar plot
                plt.bar(data_x, data_y, color="#5ca3d0")

                plt.xlabel("Parameter")
                plt.ylabel("importance percentage")
                plt.xticks(rotation=20)

                plt.title("Parameter Importance")
                # plt.show()

                pic_IObytes = io.BytesIO()
                fig.savefig(pic_IObytes, format="png", dpi=400)
                pic_IObytes.seek(0)
                pic_hash = base64.b64encode(pic_IObytes.read())
                plt.close(fig)
                model_report += "<h3>Parameter Importance</h3>"
                model_report += (
                    """
                    <p><img style="width:750px" alt="AutoDL Performance" src="data:image/png;base64,"""
                    + pic_hash.decode("utf-8")
                    + """" /></p>
                    """
                )
            except:
                pass

        if not save_to_folder:
            model_report += "<hr/><h2>Best Performing Model Report</h2>"
            folder_name = (
                "AutoDL_"
                + df["Model"][0]
                + "_"
                + df["backbone"][0]
                + "_"
                + df["timing"][0]
            )
            html_report = os.path.join(
                output_folder, "models", folder_name, "model_metrics.html"
            )

            model_metrics = open(html_report, "r")
            model_metrics_content = model_metrics.readlines()
            model_metrics.close()
            model_report += "\n".join(model_metrics_content)
        if save_to_folder:
            model_report += "<hr/><h2>Best Performing Model Report</h2>"
            html_report = os.path.join(output_folder, "model_metrics.html")
            model_metrics = open(html_report, "r")
            model_metrics_content = model_metrics.readlines()
            model_metrics.close()
            model_report += "\n".join(model_metrics_content)

        ## Ends
        # if not save_evaluated_models:
        #     model_report += "<br/><br/><p>* To display full model metrics enable Save Evaluated Models parameter.</p>"
        content += model_report

        content += "</div>"

        ## Hidden divs
        exhaustive_dfs = exhaustive_mode_studies

        for i, ex in enumerate(exhaustive_dfs):
            hidden_divs = ""
            name = ex.study_name
            best_params = ex.best_params
            ex_df = ex.trials_dataframe().sort_values("value", ascending=False)
            del ex_df["number"]
            del ex_df["datetime_start"]
            del ex_df["datetime_complete"]
            del ex_df["state"]
            for ind, dl in ex_df.iterrows():
                ids = name + "_" + dl["params_backbones"]
                hidden_divs += (
                    '<div id="'
                    + str(ids)
                    + str(round(dl["value"], 6)).replace(".", "_")
                    + '" style="display: none">'
                )
                hidden_divs += "<h1>Summary of " + str(name) + "</h1>"
                hidden_divs += (
                    """<p><a onclick="toggleShow(\'"""
                    + str(ids)
                    + str(round(dl["value"], 6)).replace(".", "_")
                    + """\');toggleShow('main')" >&lt;&lt; Go back</a></p>"""
                )
                hidden_divs += "<h2>Model name: <u>" + str(name) + "</u> </h2>"
                hidden_divs += "<h2>Model parameters</h2>"
                hidden_divs += "<ul>"
                for key, val in dl.items():
                    hidden_divs += (
                        "<li><strong>" + str(key) + "</strong>: " + str(val) + "</li>"
                    )

                hidden_divs += "</ul>"
                if save_evaluated_models:
                    hidden_divs += "<h2>Model Performance Report</h2>"
                    folder_name = (
                        "AutoDL_"
                        + name
                        + "_"
                        + dl["params_backbones"]
                        + "_"
                        + ex._timing[ind]
                    )
                    html_model_report = os.path.join(
                        output_folder, "models", folder_name, "model_metrics.html"
                    )
                    model_metrics = open(html_model_report, "r")
                    model_metrics_content = model_metrics.readlines()
                    model_metrics.close()
                    hidden_divs += "\n".join(model_metrics_content)

                if not save_evaluated_models:
                    hidden_divs += (
                        "<p>*To see model metrics enable Save Evaluated Models</p>"
                    )

                hidden_divs += "</div>"
            content += hidden_divs
        ## Hidden divs ends
    else:
        model_report = ""
        if not save_to_folder:
            model_report += "<hr/><h2>Best Performing Model Report</h2>"
            folder_name = (
                "AutoDL_"
                + df["Model"][0]
                + "_"
                + df["backbone"][0]
                + "_"
                + df["timing"][0]
            )
            html_report = os.path.join(
                output_folder, "models", folder_name, "model_metrics.html"
            )

            model_metrics = open(html_report, "r")
            model_metrics_content = model_metrics.readlines()
            model_metrics.close()
            model_report += "\n".join(model_metrics_content)
        if save_to_folder:
            model_report += "<hr/><h2>Best Performing Model Report</h2>"
            html_report = os.path.join(output_folder, "model_metrics.html")
            model_metrics = open(html_report, "r")
            model_metrics_content = model_metrics.readlines()
            model_metrics.close()
            model_report += "\n".join(model_metrics_content)
        content += model_report
        content += "</div>"
    display_html_path = os.path.join(output_folder, "README.html")

    html_template = header + content + footer

    f = open(display_html_path, "w")
    # html_template = header + content + hidden_divs + footer

    # writing the code into the file
    f.write(html_template)

    # close the file
    f.close()

    return html_template


def _get_model(model, **params):
    global self_obj
    self = self_obj
    return getattr(ag.learn, model)(self._data, **params)


def _objective(trial):
    global self_obj, temp_log_msg
    all_networks = self_obj._model_stats()
    params = {}
    for key, val in all_networks[trial.study.study_name]["params"].items():
        if key == "type_list":
            for k, v in val.items():
                params[k] = trial.suggest_categorical(k, v)
        if key == "type_float":
            for k, v in val.items():
                params[k] = trial.suggest_float(k, v[0], v[1])
        if key == "type_int":
            for k, v in val.items():
                params[k] = trial.suggest_int(k, v[0], v[1])

    if self_obj.verbose:
        log_msg = "{date}: selected params: {params}".format(
            date=dt.now().strftime("%d-%m-%Y %H:%M:%S"),
            params=params,
        )
        temp_log_msg.append(log_msg)

    # Get pretrained model
    model = _get_model(trial.study.study_name, **params)

    global all_val_losses, all_train_losses, dice, BestPerformingModel, best_backbone, best_model
    callbacks = [
        self_obj._train_callback(
            model.learn,
            self_obj._tiles_required // self_obj._data.batch_size,
        )
    ]
    model.fit(
        15,
        lr=trial.suggest_float("lr", 1e-4, 1e-2),
        early_stopping=True,
        callbacks=callbacks,
        checkpoint=False,
    )
    all_val_losses.append(model.learn.recorder.get_state()["val_losses"][-1])
    all_train_losses.append(float(model.learn.recorder.get_state()["losses"][-1]))
    if self_obj._dataset_type == "classification":
        dice.append(np.array(model.learn.recorder.get_state()["metrics"])[-1][1])
        accuracy = model.accuracy()
    else:
        avg_precision = model.average_precision_score()
        accuracy = sum(avg_precision.values()) / len(avg_precision.values())

    if accuracy >= self_obj._max_accuracy:
        self_obj._is_best = True
        self_obj._max_accuracy = accuracy
        self_obj.best_model = trial.study.study_name
        self_obj._best_backbone = params["backbones"]
        self_obj.BestPerformingModel = model
    best_name_time = time.strftime("%Y-%m-%d_%H-%M-%S")

    ## Save the model
    if self_obj.verbose:
        log_msg = "{date}: Saving the model".format(
            date=dt.now().strftime("%d-%m-%Y %H:%M:%S")
        )
        print(log_msg)
        temp_log_msg.append(log_msg)

    # print(self_obj._save_evaluated_models)
    # name_time = time.strftime("%Y-%m-%d_%H-%M-%S")
    global timing

    timing.append(best_name_time)
    if self_obj._save_evaluated_models:
        if self_obj._save_to_folder:
            model.save(
                self_obj._output_path
                + os.sep
                + "models"
                + os.sep
                + "AutoDL_"
                + trial.study.study_name
                + "_"
                + params["backbones"]
                + "_"
                + best_name_time
            )
            if self_obj.verbose:
                log_msg = "{date}: model saved at {path}".format(
                    date=dt.now().strftime("%d-%m-%Y %H:%M:%S"),
                    path=os.path.join(
                        self_obj._output_path,
                        "models",
                        "AutoDL_"
                        + trial.study.study_name
                        + "_"
                        + params["backbones"]
                        + "_"
                        + best_name_time,
                    ),
                )
                print(log_msg)
                temp_log_msg.append(log_msg)
        else:
            model.save(
                "AutoDL_"
                + trial.study.study_name
                + "_"
                + params["backbones"]
                + "_"
                + best_name_time
            )
            if self_obj.verbose:
                log_msg = "{date}: model saved at {path}".format(
                    date=dt.now().strftime("%d-%m-%Y %H:%M:%S"),
                    path=os.path.join(
                        self_obj._data.path,
                        "models",
                        "AutoDL_"
                        + trial.study.study_name
                        + "_"
                        + params["backbones"]
                        + "_"
                        + best_name_time,
                    ),
                )
                print(log_msg)
                temp_log_msg.append(log_msg)

    del model
    gc.collect()
    torch.cuda.empty_cache()
    return accuracy


def _train_exhaust_mode(self, time, name, val_losses, train_losses, d):
    global temp_log_msg, self_obj, all_val_losses, all_train_losses, dice, timing, best_name_time

    self_obj = self
    timing = []
    all_val_losses, all_train_losses, dice = val_losses, train_losses, d
    temp_log_msg = []

    sampler = optuna.samplers.TPESampler()
    if self_obj.verbose:
        log_msg = "{date}: A new study created in memory with name: {networks}".format(
            date=dt.now().strftime("%d-%m-%Y %H:%M:%S"),
            networks=name,
        )
        self_obj._logger_dict.append(log_msg)
        clear_output(wait=True)
        all_logs = "\n".join(self_obj._logger_dict)
        print(all_logs)
    study = optuna.create_study(
        sampler=sampler,
        pruner=optuna.pruners.MedianPruner(
            n_startup_trials=3, n_warmup_steps=3, interval_steps=2
        ),
        direction="maximize",
        study_name=name,
    )
    study.optimize(
        func=_objective,
        n_trials=None,
        show_progress_bar=True,
        timeout=time,
        gc_after_trial=True,
    )
    if self_obj.verbose:
        self_obj._logger_dict.append("\n".join(temp_log_msg))
        clear_output(wait=True)
        all_logs = "\n".join(self_obj._logger_dict)
        print(all_logs)
    return study, self_obj, all_val_losses, all_train_losses, dice, timing


def _get_emd_path(emd_path):
    emd_path = Path(emd_path)
    if emd_path.suffix == ".dlpk":
        temp_path = _temp_dlpk(emd_path)
        emd_path = Path(temp_path)
        # return cls.from_model(temp_path)

    if emd_path.suffix != ".emd":
        list_files = get_files(emd_path, extensions=[".emd"])
        assert len(list_files) == 1
        # return cls.from_model(list_files[0])
        emd_path = list_files[0]
    return emd_path


def _temp_dlpk(dlpk_path):
    with ZipFile(dlpk_path, "r") as zip_obj:
        temp_dir = tempfile.TemporaryDirectory().name
        zip_obj.extractall(temp_dir)
    return temp_dir



from .env import HAS_TENSORFLOW

if HAS_TENSORFLOW:
    import tensorflow as tf
    from tensorflow.keras.layers import Layer
    from tensorflow.keras import applications
    import inspect
    from .._utils.fastai_tf_fit import _pytorch_to_tf_batch


## Layers Start ##
# Imagenet Stats
mean = tf.convert_to_tensor([0.485, 0.456, 0.406], dtype=tf.float32)
std = tf.convert_to_tensor([0.229, 0.224, 0.225], dtype=tf.float32)


class NormalizationLayerRGB(Layer):
    def __call__(self, input):
        x = (input + 1.0) / 2.0
        x = (x - mean) / std
        return x


class UpSample2DToSize(Layer):
    def __init__(self, output_size, name=None, **kwargs):
        super(UpSample2DToSize, self).__init__()
        if isinstance(output_size, (int, float)):
            output_size = (output_size, output_size)
        self.size = output_size

    def call(self, input):
        if tf.keras.backend.image_data_format() == "channels_first":
            x = tf.transpose(input, perm=[0, 2, 3, 1])
            x = self.resize(x)
            return tf.transpose(x, perm=[0, 3, 1, 2])
        else:
            return self.resize(input)

    def resize(self, input):
        return tf.image.resize(input, self.size)


## Layers END ##

## Backbone Functions Start ##
backbone_store = {}
for prop in dir(applications):
    prop_obj = getattr(applications, prop)
    if inspect.isfunction(prop_obj):
        backbone_store[prop.lower()] = prop_obj


def handle_backbone_parameter(backbone):
    if backbone is None:
        backbone = "ResNet50"
    if type(backbone) == str:
        backbone = get_backbone(backbone)
    return backbone


def get_backbone(backbone_name: str):
    return backbone_store[backbone_name.lower()]


_mobile_opt_backbones = ["MobileNetV2", "MobileNet"]
mobile_opt_backbones = [get_backbone(backbone) for backbone in _mobile_opt_backbones]


def check_backbone_is_mobile_optimized(backbone):
    return backbone in mobile_opt_backbones


## Backbone Functions end ##


def get_input_shape(chip_size=None, dl=None):
    if dl is not None:
        x, y = next(iter(dataloader))
        chip_size = x.shape[-1]
    if tf.keras.backend.image_data_format() == "channels_last":
        in_shape = [chip_size, chip_size, 3]
    else:
        in_shape = [3, chip_size, chip_size]
    return in_shape


def get_padding_info(op_layer, target_layer, data_format=None):
    """
    Returns the required padding info for matching two layers
    to same spatial size using ZeroPadding.

    Here op_layer is smaller than the target_layer in spatial size.
    """
    if data_format is None:
        data_format = tf.keras.backend.image_data_format()
    if data_format == "channels_last":
        _, target_rows, target_cols, _ = target_layer.shape
        _, op_rows, op_cols, _ = op_layer.shape
    else:  # channel first
        _, _, target_rows, target_cols = target_layer.shape
        _, _, op_rows, op_cols = op_layer.shape

    top_pad = (target_rows - op_rows) // 2
    left_pad = (target_cols - op_cols) // 2
    bottom_pad = (target_rows - op_rows) - top_pad
    right_pad = (target_cols - op_cols) - left_pad

    return top_pad, bottom_pad, left_pad, right_pad


def get_cropping_info(op_layer, target_layer, data_format=None):
    """
    Returns the required cropping info for matching two layers
    to same spatial size using ZeroPadding.

    Here op_layer is bigger than the target_layer in spatial size.
    """
    top_crop, bottom_crop, left_crop, right_crop = get_padding_info(
        target_layer, op_layer, data_format=data_format
    )
    return top_crop, bottom_crop, left_crop, right_crop


def get_skip_connection_info(backbone_initalized, thresh=1.5, data_format=None):
    """
    Returns information about the possible skip connections
    that can be used by any U-Net type model architecture on
    the supplied backbone.
    """
    ca = get_channel_axis(data_format=data_format)
    skip_lyr_idxs = []
    skip_lyr_channels = []
    layers = backbone_initalized.layers
    input_shape = previous_shape = layers[0].input.shape
    for i, layer in enumerate(layers):
        new_shape = layer.output.shape
        if new_shape[1] * thresh < previous_shape[1]:
            skip_lyr_idxs.insert(0, i - 1)
            skip_lyr_channels.insert(0, previous_shape[ca])
        previous_shape = new_shape
    return skip_lyr_idxs, skip_lyr_channels


def get_channel_axis(data_format=None):
    if data_format is None:
        data_format = tf.keras.backend.image_data_format()
    # Get channel axis
    if data_format == "channels_last":
        channel_axis = -1
    else:
        channel_axis = 1
    return channel_axis


def predict_batch_tf(self, imagetensor_batch):
    return self.learn.model(
        _pytorch_to_tf_batch(imagetensor_batch), training=False
    ).detach()



"""

The ``Symbology`` class generates Symbol Types for the :class:`~arcgis.gis.GIS` object.

"""
import json
import arcgis
from arcgis.features import FeatureCollection, FeatureSet
from arcgis.gis import GIS
from arcgis.geometry import _types

###########################################################################
ALLOWED_CMAPS = [
    "Accent",
    "Accent_r",
    "Blues",
    "Blues_r",
    "BrBG",
    "BrBG_r",
    "BuGn",
    "BuGn_r",
    "BuPu",
    "BuPu_r",
    "CMRmap",
    "CMRmap_r",
    "Dark2",
    "Dark2_r",
    "GnBu",
    "GnBu_r",
    "Greens",
    "Greens_r",
    "Greys",
    "Greys_r",
    "OrRd",
    "OrRd_r",
    "Oranges",
    "Oranges_r",
    "PRGn",
    "PRGn_r",
    "Paired",
    "Paired_r",
    "Pastel1",
    "Pastel1_r",
    "Pastel2",
    "Pastel2_r",
    "PiYG",
    "PiYG_r",
    "PuBu",
    "PuBuGn",
    "PuBuGn_r",
    "PuBu_r",
    "PuOr",
    "PuOr_r",
    "PuRd",
    "PuRd_r",
    "Purples",
    "Purples_r",
    "RdBu",
    "RdBu_r",
    "RdGy",
    "RdGy_r",
    "RdPu",
    "RdPu_r",
    "RdYlBu",
    "RdYlBu_r",
    "RdYlGn",
    "RdYlGn_r",
    "Reds",
    "Reds_r",
    "Set1",
    "Set1_r",
    "Set2",
    "Set2_r",
    "Set3",
    "Set3_r",
    "Spectral",
    "Spectral_r",
    "Wistia",
    "Wistia_r",
    "YlGn",
    "YlGnBu",
    "YlGnBu_r",
    "YlGn_r",
    "YlOrBr",
    "YlOrBr_r",
    "YlOrRd",
    "YlOrRd_r",
    "afmhot",
    "afmhot_r",
    "autumn",
    "autumn_r",
    "binary",
    "binary_r",
    "bone",
    "bone_r",
    "brg",
    "brg_r",
    "bwr",
    "bwr_r",
    "cool",
    "cool_r",
    "coolwarm",
    "coolwarm_r",
    "copper",
    "copper_r",
    "cubehelix",
    "cubehelix_r",
    "flag",
    "flag_r",
    "gist_earth",
    "gist_earth_r",
    "gist_gray",
    "gist_gray_r",
    "gist_heat",
    "gist_heat_r",
    "gist_ncar",
    "gist_ncar_r",
    "gist_rainbow",
    "gist_rainbow_r",
    "gist_stern",
    "gist_stern_r",
    "gist_yarg",
    "gist_yarg_r",
    "gnuplot",
    "gnuplot2",
    "gnuplot2_r",
    "gnuplot_r",
    "gray",
    "gray_r",
    "hot",
    "hot_r",
    "hsv",
    "hsv_r",
    "inferno",
    "inferno_r",
    "jet",
    "jet_r",
    "magma",
    "magma_r",
    "nipy_spectral",
    "nipy_spectral_r",
    "ocean",
    "ocean_r",
    "pink",
    "pink_r",
    "plasma",
    "plasma_r",
    "prism",
    "prism_r",
    "rainbow",
    "rainbow_r",
    "seismic",
    "seismic_r",
    "spring",
    "spring_r",
    "summer",
    "summer_r",
    "terrain",
    "terrain_r",
    "viridis",
    "viridis_r",
    "winter",
    "winter_r",
]
###########################################################################
_CMAP_LOOKUP = dict(zip([c.lower() for c in ALLOWED_CMAPS], ALLOWED_CMAPS))
###########################################################################
_SYMBOL_TYPES = {
    "simple": "Simple Point, Line or Polygon Symbol",
    "text": "Text Symbol",
    "picture": "Picture Marker or Fill Symbol",
    "carto": "Cartographic Line Symbol",
}
###########################################################################
LINE_STYLES = {
    "s": "esriSLSSolid",  # default
    "-": "esriSLSDash",
    "-.": "esriSLSDashDot",
    "-..": "esriSLSDashDotDot",
    ".": "esriSLSDot",
    "--": "esriSLSLongDash",
    "--.": "esriSLSLongDashDot",
    "n": "esriSLSNull",
    "s-": "esriSLSShortDash",
    "s-.": "esriSLSShortDashDot",
    "s-..": "esriSLSShortDashDotDot",
    "s.": "esriSLSShortDot",
}
###########################################################################
POINT_STYLES = {
    "o": "esriSMSCircle",  # default
    "+": "esriSMSCross",
    "d": "esriSMSDiamond",
    "s": "esriSMSSquare",
    "x": "esriSMSX",
    # "^" : "esriSMSTriangle" # Does not render in web maps
}
###########################################################################
POLYGON_STYLES = {
    "\\": "esriSFSBackwardDiagonal",
    "/": "esriSFSForwardDiagonal",
    "|": "esriSFSVertical",
    "-": "esriSFSHorizontal",
    "x": "esriSFSDiagonalCross",
    "+": "esriSFSCross",
    "s": "esriSFSSolid",  # default
}
###########################################################################
POLYGON_STYLES_DISPLAY = {
    "\\": "Backward Diagonal",
    "/": "Forward Diagonal",
    "|": "Vertical Bar",
    "-": "Horizontal Bar",
    "x": "Diagonal Cross",
    "+": "Cross",
    "s": "Solid Fill (default)",  # default
}
###########################################################################
POINT_STYLES_DISPLAY = {
    "o": "Circle (default)",  # default
    "+": "Cross",
    "d": "Diamond",
    "s": "Square",
    "x": "X",
    # "^" : "esriSMSTriangle" # Does not render in web maps
}
###########################################################################
LINE_STYLES_DISPLAY = {
    "s": "Solid (default)",  # default
    "-": "Dash",
    "-.": "Dash Dot",
    "-..": "Dash Dot Dot",
    ".": "Dot",
    "--": "Long Dash",
    "--.": "Long Dash Dot",
    "n": "Null",
    "s-": "Short Dash",
    "s-.": "Short Dash Dot",
    "s-..": "Short Dash Dot Dot",
    "s.": "Short Dot",
}
###########################################################################
# Have colormaps separated into categories:
# http://matplotlib.org/examples/color/colormaps_reference.html
cmaps = [
    ("Perceptually Uniform Sequential", ["viridis", "plasma", "inferno", "magma"]),
    (
        "Sequential",
        [
            "Greys",
            "Purples",
            "Blues",
            "Greens",
            "Oranges",
            "Reds",
            "YlOrBr",
            "YlOrRd",
            "OrRd",
            "PuRd",
            "RdPu",
            "BuPu",
            "GnBu",
            "PuBu",
            "YlGnBu",
            "PuBuGn",
            "BuGn",
            "YlGn",
        ],
    ),
    (
        "Sequential (2)",
        [
            "binary",
            "gist_yarg",
            "gist_gray",
            "gray",
            "bone",
            "pink",
            "spring",
            "summer",
            "autumn",
            "winter",
            "cool",
            "Wistia",
            "hot",
            "afmhot",
            "gist_heat",
            "copper",
        ],
    ),
    (
        "Diverging",
        [
            "PiYG",
            "PRGn",
            "BrBG",
            "PuOr",
            "RdGy",
            "RdBu",
            "RdYlBu",
            "RdYlGn",
            "Spectral",
            "coolwarm",
            "bwr",
            "seismic",
        ],
    ),
    (
        "Qualitative",
        [
            "Pastel1",
            "Pastel2",
            "Paired",
            "Accent",
            "Dark2",
            "Set1",
            "Set2",
            "Set3",
            "tab10",
            "tab20",
            "tab20b",
            "tab20c",
        ],
    ),
    (
        "Miscellaneous",
        [
            "flag",
            "prism",
            "ocean",
            "gist_earth",
            "terrain",
            "gist_stern",
            "gnuplot",
            "gnuplot2",
            "CMRmap",
            "cubehelix",
            "brg",
            "hsv",
            "gist_rainbow",
            "rainbow",
            "jet",
            "nipy_spectral",
            "gist_ncar",
        ],
    ),
]


###########################################################################
def _cmap2rgb(colors, step, alpha=1):
    """converts a color map to RGBA list"""
    from matplotlib import cm

    t = getattr(cm, colors)(step, bytes=True)
    t = [int(i) for i in t]
    t[-1] = alpha * 255
    return t


###########################################################################
def display_colormaps(colors=None):
    """
    The ``display_colormaps`` method displays a visual colormaps in order to assist users in selecting a color scheme
    for the data they wish to display on a map, or in a web map.

    .. note::
        ``display_colormaps is a variation of the
        `colormaps reference <http://matplotlib.org/examples/color/colormaps_reference.html>`_ page for `matplotlib`.

    """
    import numpy as np
    import matplotlib.pyplot as plt

    if colors is None:
        cmaps = [("All Color Maps", ALLOWED_CMAPS)]
    elif isinstance(colors, (str)):
        cmaps = [("Selected Color Map", [_CMAP_LOOKUP[colors.lower()]])]
    elif isinstance(colors, (tuple, list)):
        colors = [_CMAP_LOOKUP[color.lower()] for color in colors]
        cmaps = [("Selected Color Map", colors)]
    nrows = max(len(cmap_list) for cmap_category, cmap_list in cmaps)
    gradient = np.linspace(0, 1, 256)
    gradient = np.vstack((gradient, gradient))

    def plot_color_gradients(cmap_category, cmap_list, nrows):
        """creates the entry for the plot"""
        if nrows == 1:
            nrows = 2
        height = 0.2 * nrows
        fig, axes = plt.subplots(nrows=nrows, figsize=(6, height))
        fig.subplots_adjust(top=0.95, bottom=0.01, left=0.2, right=0.99)
        axes[0].set_title(cmap_category, fontsize=12)

        for ax, name in zip(axes, cmap_list):
            ax.imshow(gradient, aspect="auto", cmap=plt.get_cmap(name))
            pos = list(ax.get_position().bounds)
            x_text = pos[0] - 0.01
            y_text = pos[1] + pos[3] / 2.0
            fig.text(x_text, y_text, name, va="center", ha="right", fontsize=10)

        # Turn off *all* ticks & spines, not just the ones with colormaps.
        for ax in axes:
            ax.set_axis_off()

    for cmap_category, cmap_list in cmaps:
        plot_color_gradients(cmap_category, cmap_list, nrows)

    plt.show()


###########################################################################
def show_styles(geometry_type):
    """
    The ``show_styles`` method retrieves the available styles for a given geometry type as a Pandas dataframe.

    :return:
        A Pandas Dataframe
    """
    import pandas as pd

    if geometry_type.lower() in ["point", "multipoint"]:
        c = POINT_STYLES_DISPLAY
    elif geometry_type.lower() in ["polyline", "line"]:
        c = LINE_STYLES_DISPLAY
    elif geometry_type.lower() in ["polygon"]:
        c = POLYGON_STYLES_DISPLAY
    else:
        raise Exception("Invalid geometry type")

    d = [(k, v) for k, v in c.items()]
    return pd.DataFrame(data=d, columns=["MARKER", "ESRI_STYLE"])


###########################################################################
def create_symbol(
    geometry_type, symbol_type=None, symbol_style=None, colors=None, **kwargs
):
    """
    The ``create_symbol`` method generates a Symbol from a given set of parameters.

    ``create_symbol`` creates either a ``Picture``, ``Text``, ``Cartographic``, or ``Simple Symbol``
    based on a given set of parameters.

    .. note::
        Each symbol type has a specific set of
        parameters that are excepted.  There is a simplified input definition similar to `matplotlib`.


    =======================  =========================================================
    **Required Argument**    **Description**
    -----------------------  ---------------------------------------------------------
    geometry_type            required string.  This can be ``point``, ``line``, ``polygon``, or
                             ``multipoint``. It helps to ensure that the symbol created
                             will be supported by that geometry type.
    =======================  =========================================================


    =======================  =========================================================
    **Optional Argument**    **Description**
    -----------------------  ---------------------------------------------------------
    symbol_type              optional string. This is the type of symbol the user
                             needs to create.  Valid inputs are: simple, picture, text,
                             or carto.  The default is simple.
    -----------------------  ---------------------------------------------------------
    symbol_style             optional string. This is the symbology used by the
                             geometry.  For example 's' for a Line geometry is a solid
                             line. And '-' is a dash line.

                             **Point Symbols**

                             + 'o' - Circle (default)
                             + '+' - Cross
                             + 'D' - Diamond
                             + 's' - Square
                             + 'x' - X

                             **Polyline Symbols**

                             + 's' - Solid (default)
                             + '-' - Dash
                             + '-.' - Dash Dot
                             + '-..' - Dash Dot Dot
                             + '.' - Dot
                             + '--' - Long Dash
                             + '--.' - Long Dash Dot
                             + 'n' - Null
                             + 's-' - Short Dash
                             + 's-.' - Short Dash Dot
                             + 's-..' - Short Dash Dot Dot
                             + 's.' - Short Dot

                             **Polygon Symbols**

                             + 's' - Solid Fill (default)
                             + '\' - Backward Diagonal
                             + '/' - Forward Diagonal
                             + '|' - Vertical Bar
                             + '-' - Horizontal Bar
                             + 'x' - Diagonal Cross
                             + '+' - Cross
    -----------------------  ---------------------------------------------------------
    colors                   Optional string or list.  This is the color scheme a user can provide if the exact color is not needed, or a user can provide a list with the color defined as: [red, green blue, alpha]. The values red, green, blue are from 0-255 and alpha is a float value from 0 - 1. The default value is 'jet' color scheme.
    -----------------------  ---------------------------------------------------------
    cstep                    optional integer.  If provided, it's the color location on
                             the color scheme.
    =======================  =========================================================

    **Simple Symbols**

    This is a list of optional parameters that can be given for point, line or
    polygon geometries.

    ====================  =========================================================
    **Parameter**          **Description**
    --------------------  ---------------------------------------------------------
    marker_size           optional float.  Numeric size of the symbol given in
                          points.
    --------------------  ---------------------------------------------------------
    marker_angle          optional float. Numeric value used to rotate the symbol.
                          The symbol is rotated counter-clockwise. For example,
                          The following, angle=-30, in will create a symbol rotated
                          -30 degrees counter-clockwise; that is, 30 degrees
                          clockwise.
    --------------------  ---------------------------------------------------------
    marker_xoffset        Numeric value indicating the offset on the x-axis in points.
    --------------------  ---------------------------------------------------------
    marker_yoffset        Numeric value indicating the offset on the y-axis in points.
    --------------------  ---------------------------------------------------------
    line_width            optional float. Numeric value indicating the width of the line in points
    --------------------  ---------------------------------------------------------
    outline_style         Optional string. For polygon point, and line geometries , a
                          customized outline type can be provided.

                          Allowed Styles:

                          + 's' - Solid (default)
                          + '-' - Dash
                          + '-.' - Dash Dot
                          + '-..' - Dash Dot Dot
                          + '.' - Dot
                          + '--' - Long Dash
                          + '--.' - Long Dash Dot
                          + 'n' - Null
                          + 's-' - Short Dash
                          + 's-.' - Short Dash Dot
                          + 's-..' - Short Dash Dot Dot
                          + 's.' - Short Dot
    --------------------  ---------------------------------------------------------
    outline_color         optional string or list.  This is the same color as the
                          colors property, but specifically applies to the outline_color.
    ====================  =========================================================

    **Picture Symbol**

    This type of symbol only applies to Points, MultiPoints and Polygons.

    ====================  =========================================================
    **Parameter**          **Description**
    --------------------  ---------------------------------------------------------
    marker_angle          Numeric value that defines the number of degrees ranging
                          from 0-360, that a marker symbol is rotated. The rotation
                          is from East in a counter-clockwise direction where East
                          is the 0 axis.
    --------------------  ---------------------------------------------------------
    marker_xoffset        Numeric value indicating the offset on the x-axis in points.
    --------------------  ---------------------------------------------------------
    marker_yoffset        Numeric value indicating the offset on the y-axis in points.
    --------------------  ---------------------------------------------------------
    height                Numeric value used if needing to resize the symbol. Specify a value in points. If images are to be displayed in their original size, leave this blank.
    --------------------  ---------------------------------------------------------
    width                 Numeric value used if needing to resize the symbol. Specify a value in points. If images are to be displayed in their original size, leave this blank.
    --------------------  ---------------------------------------------------------
    url                   String value indicating the URL of the image. The URL should be relative if working with static layers. A full URL should be used for map service dynamic layers. A relative URL can be dereferenced by accessing the map layer image resource or the feature layer image resource.
    --------------------  ---------------------------------------------------------
    image_data            String value indicating the base64 encoded data.
    --------------------  ---------------------------------------------------------
    xscale                Numeric value indicating the scale factor in x direction.
    --------------------  ---------------------------------------------------------
    yscale                Numeric value indicating the scale factor in y direction.
    --------------------  ---------------------------------------------------------
    outline_color         optional string or list.  This is the same color as the
                          colors property, but specifically applies to the outline_color.
    --------------------  ---------------------------------------------------------
    outline_style         Optional string. For polygon point, and line geometries , a
                          customized outline type can be provided.

                          Allowed Styles:

                          + 's' - Solid (default)
                          + '-' - Dash
                          + '-.' - Dash Dot
                          + '-..' - Dash Dot Dot
                          + '.' - Dot
                          + '--' - Long Dash
                          + '--.' - Long Dash Dot
                          + 'n' - Null
                          + 's-' - Short Dash
                          + 's-.' - Short Dash Dot
                          + 's-..' - Short Dash Dot Dot
                          + 's.' - Short Dot
    --------------------  ---------------------------------------------------------
    outline_color         optional string or list.  This is the same color as the
                          colors property, but specifically applies to the outline_color.
    --------------------  ---------------------------------------------------------
    line_width            optional float. Numeric value indicating the width of the line in points
    ====================  =========================================================

    **Text Symbol**

    This type of symbol only applies to Points, MultiPoints and Polygons.

    ====================  =========================================================
    **Parameter**          **Description**
    --------------------  ---------------------------------------------------------
    font_decoration       The text decoration. Must be one of the following values:

                          - line-through
                          - underline
                          - none
    --------------------  ---------------------------------------------------------
    font_family           Optional string. The font family.
    --------------------  ---------------------------------------------------------
    font_size             Optional float. The font size in points.
    --------------------  ---------------------------------------------------------
    font_style            Optional string. The text style.

                          - italic
                          - normal
                          - oblique
    --------------------  ---------------------------------------------------------
    font_weight           Optional string. The text weight.
                          Must be one of the following values:

                          - bold
                          - bolder
                          - lighter
                          - normal
    --------------------  ---------------------------------------------------------
    background_color      optional string/list. Background color is represented as
                          a four-element array or string of a color map.
    --------------------  ---------------------------------------------------------
    halo_color            Optional string/list. Color of the halo around the text.
                          The default is None.
    --------------------  ---------------------------------------------------------
    halo_size             Optional integer/float. The point size of a halo around
                          the text symbol.
    --------------------  ---------------------------------------------------------
    horizontal_alignment  optional string. One of the following string values
                          representing the horizontal alignment of the text.
                          Must be one of the following values:

                          - left
                          - right
                          - center
                          - justify
    --------------------  ---------------------------------------------------------
    kerning               optional boolean. Boolean value indicating whether to
                          adjust the spacing between characters in the text string.
    --------------------  ---------------------------------------------------------
    line_color            optional string/list. Outline color is represented as
                          a four-element array or string of a color map.
    --------------------  ---------------------------------------------------------
    line_width            optional integer/float. Outline size.
    --------------------  ---------------------------------------------------------
    marker_angle          optional int. A numeric value that defines the number of
                          degrees (0 to 360) that a text symbol is rotated. The
                          rotation is from East in a counter-clockwise direction
                          where East is the 0 axis.
    --------------------  ---------------------------------------------------------
    marker_xoffset        optional int/float.Numeric value indicating the offset
                          on the x-axis in points.
    --------------------  ---------------------------------------------------------
    marker_yoffset        optional int/float.Numeric value indicating the offset
                          on the x-axis in points.
    --------------------  ---------------------------------------------------------
    right_to_left         optional boolean. Set to true if using Hebrew or Arabic
                          fonts.
    --------------------  ---------------------------------------------------------
    rotated               optional boolean. Boolean value indicating whether every
                          character in the text string is rotated.
    --------------------  ---------------------------------------------------------
    text                  Required string.  Text Value to display next to geometry.
    --------------------  ---------------------------------------------------------
    vertical_alignment    Optional string. One of the following string values
                          representing the vertical alignment of the text.
                          Must be one of the following values:

                          - top
                          - bottom
                          - middle
                          - baseline
    ====================  =========================================================

    **Cartographic Symbol**

    This type of symbol only applies to line geometries.

    ====================  =========================================================
    **Parameter**          **Description**
    --------------------  ---------------------------------------------------------
    line_width            optional float. Numeric value indicating the width of the line in points
    --------------------  ---------------------------------------------------------
    cap                   Optional string.  The cap style.
    --------------------  ---------------------------------------------------------
    join                  Optional string. The join style.
    --------------------  ---------------------------------------------------------
    miter_limit           Optional string. Size threshold for showing mitered line joins.
    ====================  =========================================================

    :return: Dictionary

    """
    import numpy as np
    import matplotlib.pyplot as plt

    alpha = kwargs.pop("alpha", 1)
    symbol = kwargs.pop("symbol", None)
    if symbol_type is None:
        symbol_type = "simple"

    gtype = geometry_type.upper()
    if colors is None:
        colors = "jet"

    # Get color step, if not specified, pick a random value between 0 and 255.
    cstep = kwargs.pop("cstep", int(np.random.randint(0, 255)))
    if cstep is None:
        cstep = int(np.random.randint(0, 255))
    renderer_type = "simple"

    marker_size = kwargs.pop("marker_size", 8)
    marker_angle = kwargs.pop("marker_angle", 0)
    marker_xoffset = kwargs.pop("marker_xoffset", 0)
    marker_yoffset = kwargs.pop("marker_yoffset", 0)

    line_width = kwargs.pop("line_width", 2)

    outline_style = LINE_STYLES[kwargs.pop("outline_style", "s")]

    # get outline_color if specified, else use a nice mild gray
    outline_color = kwargs.pop("outline_color", [128, 128, 128, 255])

    # get a random color if a colormap is specified
    if isinstance(colors, str):
        colors = list(_cmap2rgb(colors=colors, step=cstep, alpha=alpha))

    # if specific 4 color tuple is specified, go with it
    elif isinstance(colors, (list, tuple)) and len(colors) == 4:
        colors = colors

    # if not, the first color in the 'jet' color map. This is a deep blue
    else:
        colors = list(_cmap2rgb(colors="jet", step=cstep, alpha=alpha))

    if isinstance(outline_color, str):
        outline_cmap = list(
            _cmap2rgb(
                colors=outline_color, step=int(np.random.randint(0, 255)), alpha=alpha
            )
        )
    elif isinstance(outline_color, (tuple, list)) and len(outline_color) == 4:
        outline_cmap = outline_color
    else:
        outline_cmap = list(
            _cmap2rgb(colors="jet", step=int(np.random.randint(0, 255)), alpha=alpha)
        )

    if symbol is not None:
        return symbol

    if symbol_type.lower() == "simple":  # Default Simple Symbol
        if symbol_style is None:
            if gtype == "POINT":
                symbol_type = "esriSMS"
                symbol_style = POINT_STYLES["o"]
            elif gtype in ["POLYLINE", "LINE"]:
                symbol_type = "esriSLS"
                symbol_style = LINE_STYLES["s"]
            elif gtype in ["POLYGON"]:
                symbol_type = "esriSFS"
                symbol_style = POLYGON_STYLES["s"]
            else:
                raise Exception(
                    "Invalid geometry types only points, lines, and polygons can be plotted"
                )
        else:
            if gtype == "POINT":
                symbol_type = "esriSMS"
                symbol_style = POINT_STYLES[symbol_style]
            elif gtype in ["POLYLINE", "LINE"]:
                symbol_type = "esriSLS"
                symbol_style = LINE_STYLES[symbol_style]
            elif gtype in ["POLYGON"]:
                symbol_type = "esriSFS"
                symbol_style = POLYGON_STYLES[symbol_style]
            else:
                raise Exception(
                    "Invalid geometry types only points, lines, and polygons can be plotted"
                )
        # build the symbol
        symbol = dict()
        symbol["type"] = symbol_type
        if isinstance(colors, str):
            colors = list(_cmap2rgb(colors=colors, step=cstep, alpha=alpha))
        elif isinstance(colors, (list, tuple)) and len(colors) == 4:
            colors = colors
        else:
            colors = list(_cmap2rgb(colors="jet", step=cstep, alpha=alpha))
        if isinstance(outline_color, str):
            outline_cmap = list(
                _cmap2rgb(
                    colors=outline_color,
                    step=int(np.random.randint(0, 255)),
                    alpha=alpha,
                )
            )
        elif isinstance(outline_color, (tuple, list)) and len(outline_color) == 4:
            outline_cmap = outline_color
        else:
            outline_cmap = list(
                _cmap2rgb(
                    colors="jet", step=int(np.random.randint(0, 255)), alpha=alpha
                )
            )
        if gtype == "POINT":
            symbol["style"] = symbol_style
            symbol["color"] = colors
            symbol["size"] = marker_size
            symbol["angle"] = marker_angle
            symbol["xoffset"] = marker_xoffset
            symbol["yoffset"] = marker_yoffset
            symbol["outline"] = {
                "style": outline_style,
                "type": "esriSLS",
                "color": outline_cmap,
                "width": line_width,
            }
        elif gtype in ["LINE", "POLYLINE"]:
            symbol["style"] = symbol_style
            symbol["color"] = colors
            symbol["width"] = line_width
        elif gtype in ["POLYGON"]:
            symbol["style"] = symbol_style
            symbol["color"] = colors
            symbol["outline"] = {
                "style": outline_style,
                "type": "esriSLS",
                "color": outline_cmap,
                "width": line_width,
            }
        return symbol
    elif symbol_type == "picture" and geometry_type.lower() in [
        "multipoint",
        "point",
        "polygon",
    ]:
        if geometry_type.lower() in ["point", "multipoint"]:
            symbol = {
                "type": "esriPMS",
                "angle": marker_angle,
                "xoffset": marker_xoffset,
                "yoffset": marker_yoffset,
                "width": kwargs.pop("width", 10),
                "height": kwargs.pop("height", 10),
                "url": kwargs.pop("url", ""),
                "imageData": kwargs.pop("image_data", ""),
            }
            return symbol
        elif geometry_type.lower() == "polygon":
            symbol = {
                "type": "esriPFS",
                "color": colors,
                "xoffset": marker_xoffset,
                "yoffset": marker_yoffset,
                "xscale": kwargs.pop("xscale", None),
                "yscale": kwargs.pop("yscale", None),
                "width": kwargs.pop("width", 10),
                "height": kwargs.pop("height", 10),
                "angle": marker_angle,
                "url": kwargs.pop("url", ""),
                "imageData": kwargs.pop("image_data", None),
                "outline": {
                    "color": outline_cmap,
                    "width": line_width,
                    "type": "esriSLS",
                    "style": outline_style,
                },
            }
            return symbol
    elif symbol_type.lower() == "picture" and geometry_type.lower() in ["line"]:
        raise Exception("Line geometries do not support picture symbology")
    elif symbol_type.lower() == "text" and geometry_type.lower() in ["point"]:
        symbol = {
            "color": colors,
            "type": "esriTS",
            "horizontalAlignment": kwargs.pop("horizontal_alignment", "center"),
            "verticalAlignment": kwargs.pop("vertical_alignment", "middle"),
            "rightToLeft": kwargs.pop("right_to_left", False),
            "backgroundColor": kwargs.pop("background_color", colors),
            "borderLineColor": outline_cmap,
            "borderLineSize": line_width,
            "haloColor": kwargs.pop("halo_color", None),
            "haloSize": kwargs.pop("halo_size", None),
            "angle": marker_angle,
            "xoffset": marker_xoffset,
            "yoffset": marker_yoffset,
            "text": kwargs.pop("text", "N/A"),
            "rotated": kwargs.pop("rotated", False),
            "kerning": kwargs.pop("kerning", True),
            "font": {
                "size": kwargs.pop("font_size", 10),
                "style": kwargs.pop("font_style", "normal"),
                "weight": kwargs.pop("font_weight", "normal"),
                "family": kwargs.pop("font_familty", "serif"),
                "decoration": kwargs.pop("font_decoration", "none"),
            },
        }
        return symbol
    elif symbol_type.lower() == "text" and geometry_type.lower() not in [
        "point",
        "multipoint",
    ]:
        raise Exception("Text symbology is only supported by points and multipoints")
    elif symbol_type.lower() == "carto" and geometry_type in ["line", "polyline"]:
        cap = kwargs.pop("cap", "esriLCButt")
        join = kwargs.pop("join", "esriLJSMiter")
        miter_limit = kwargs.pop("miter_limit", 1)
        symbol = {
            "color": colors,
            "width": line_width,
            "type": "esriCLS",
            "cap": cap,
            "join": join,
            "miter_limit": miter_limit,
        }
    elif symbol_type.lower() == "carto" and geometry_type not in ["line"]:
        raise Exception("Cartographic symbology is only supported for line geomtries")
    else:
        raise Exception("Invalid symbol and geometry type")
    return symbol



from arcgis import GIS
from typing import Optional, Dict, Union, List

from ._realtime_analytics import RealTimeAnalytics
from ._util import _Util
import logging

_LOGGER = logging.getLogger(__name__)


class RealTimeAnalyticsManager:
    """
    Used to manage real-time analytic items.

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
      url                    URL of the ArcGIS Velocity organization.
    ------------------     --------------------------------------------------------------------
      gis                    an authenticated :class:`arcigs.gis.GIS` object.
    ==================     ====================================================================

    """

    _gis = None
    _util = None

    def __init__(self, url: str, gis: GIS):
        self._gis = gis

        self._util = _Util(gis, url)

    @property
    def items(self) -> List[RealTimeAnalytics]:
        """
        Get all real-time analytic items.

        :return: returns a collection of all real-time analytics items with id and label.

        .. code-block:: python

            # Get all real-time analytics items

            all_realtime_analytics = realtime_analytics.items
            all_realtime_analytics

        """
        all_realtime_analytics_response = self._util._get_request("analytics/realtime")

        if (
            all_realtime_analytics_response is not None
            and type(all_realtime_analytics_response) is list
        ):
            realtime_analytics_items = [
                RealTimeAnalytics(self._gis, self._util, realtime_item)
                for realtime_item in all_realtime_analytics_response
            ]
            return realtime_analytics_items
        elif all_realtime_analytics_response is None:
            _LOGGER.warning("No Real-time analytic items found for the user.")
            return []
        else:
            raise Exception(
                f"Error retrieving Real-time analytic items. Velocity response: ${all_realtime_analytics_response}"
            )

    def get(self, id) -> RealTimeAnalytics:
        """
        Get real-time analytic item by ID.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
          id                     Unique ID of a real-time analytic.
        ==================     ====================================================================

        :return: endpoint response of real-time analytics for the given id and label

        .. code-block:: python

            # Get real-time analytics by id
            # Method: <item>.get(id)

            sample_realtime_task = realtime_analytics.get("id")

        """
        realtime_analytics_item = self._util._get("analytics/realtime", id)
        return RealTimeAnalytics(self._gis, self._util, realtime_analytics_item)
